from __future__ import absolute_import, division

import sys
from copy import deepcopy
from math import exp, log, sqrt

import numpy as np
from numpy import dot, eye, isscalar, shape, zeros

from .common import pretty_str, reshape_z
from .stats import logpdf


class KalmanFilter(object):
    r"""Implements a Kalman filter. You are responsible for setting the
    various state variables to reasonable values; the defaults  will
    not give you a functional filter.
    For now the best documentation is my free book Kalman and Bayesian
    Filters in Python [2]_. The test files in this directory also give you a
    basic idea of use, albeit without much description.
    In brief, you will first construct this object, specifying the size of
    the state vector with dim_x and the size of the measurement vector that
    you will be using with dim_z. These are mostly used to perform size checks
    when you assign values to the various matrices. For example, if you
    specified dim_z=2 and then try to assign a 3x3 matrix to R (the
    measurement noise matrix you will get an assert exception because R
    should be 2x2. (If for whatever reason you need to alter the size of
    things midstream just use the underscore version of the matrices to
    assign directly: your_filter._R = a_3x3_matrix.)
    After construction the filter will have default matrices created for you,
    but you must specify the values for each. It’s usually easiest to just
    overwrite them rather than assign to each element yourself. This will be
    clearer in the example below. All are of type numpy.array.
    Examples
    --------
    Here is a filter that tracks position and velocity using a sensor that only
    reads position.
    First construct the object with the required dimensionality. Here the state
    (`dim_x`) has 2 coefficients (position and velocity), and the measurement
    (`dim_z`) has one. In FilterPy `x` is the state, `z` is the measurement.
    .. code::
        from filterpy.kalman import KalmanFilter
        f = KalmanFilter (dim_x=2, dim_z=1)
    Assign the initial value for the state (position and velocity). You can do this
    with a two dimensional array like so:
        .. code::
            f.x = np.array([[2.],    # position
                            [0.]])   # velocity
    or just use a one dimensional array, which I prefer doing.
    .. code::
        f.x = np.array([2., 0.])
    Define the state transition matrix:
        .. code::
            f.F = np.array([[1.,1.],
                            [0.,1.]])
    Define the measurement function. Here we need to convert a position-velocity
    vector into just a position vector, so we use:
        .. code::
        f.H = np.array([[1., 0.]])
    Define the state's covariance matrix P.
    .. code::
        f.P = np.array([[1000.,    0.],
                        [   0., 1000.] ])
    Now assign the measurement noise. Here the dimension is 1x1, so I can
    use a scalar
    .. code::
        f.R = 5
    I could have done this instead:
    .. code::
        f.R = np.array([[5.]])
    Note that this must be a 2 dimensional array.
    Finally, I will assign the process noise. Here I will take advantage of
    another FilterPy library function:
    .. code::
        from filterpy.common import Q_discrete_white_noise
        f.Q = Q_discrete_white_noise(dim=2, dt=0.1, var=0.13)
    Now just perform the standard predict/update loop:
    .. code::
        while some_condition_is_true:
            z = get_sensor_reading()
            f.predict()
            f.update(z)
            do_something_with_estimate (f.x)
    **Procedural Form**
    This module also contains stand alone functions to perform Kalman filtering.
    Use these if you are not a fan of objects.
    **Example**
    .. code::
        while True:
            z, R = read_sensor()
            x, P = predict(x, P, F, Q)
            x, P = update(x, P, z, R, H)
    See my book Kalman and Bayesian Filters in Python [2]_.
    You will have to set the following attributes after constructing this
    object for the filter to perform properly. Please note that there are
    various checks in place to ensure that you have made everything the
    'correct' size. However, it is possible to provide incorrectly sized
    arrays such that the linear algebra can not perform an operation.
    It can also fail silently - you can end up with matrices of a size that
    allows the linear algebra to work, but are the wrong shape for the problem
    you are trying to solve.
    Parameters
    ----------
    dim_x : int
        Number of state variables for the Kalman filter. For example, if
        you are tracking the position and velocity of an object in two
        dimensions, dim_x would be 4.
        This is used to set the default size of P, Q, and u
    dim_z : int
        Number of of measurement inputs. For example, if the sensor
        provides you with position in (x,y), dim_z would be 2.
    dim_u : int (optional)
        size of the control input, if it is being used.
        Default value of 0 indicates it is not used.
    compute_log_likelihood : bool (default = True)
        Computes log likelihood by default, but this can be a slow
        computation, so if you never use it you can turn this computation
        off.
    Attributes
    ----------
    x : numpy.array(dim_x, 1)
        Current state estimate. Any call to update() or predict() updates
        this variable.
    P : numpy.array(dim_x, dim_x)
        Current state covariance matrix. Any call to update() or predict()
        updates this variable.
    x_prior : numpy.array(dim_x, 1)
        Prior (predicted) state estimate. The *_prior and *_post attributes
        are for convenience; they store the  prior and posterior of the
        current epoch. Read Only.
    P_prior : numpy.array(dim_x, dim_x)
        Prior (predicted) state covariance matrix. Read Only.
    x_post : numpy.array(dim_x, 1)
        Posterior (updated) state estimate. Read Only.
    P_post : numpy.array(dim_x, dim_x)
        Posterior (updated) state covariance matrix. Read Only.
    z : numpy.array
        Last measurement used in update(). Read only.
    R : numpy.array(dim_z, dim_z)
        Measurement noise covariance matrix. Also known as the
        observation covariance.
    Q : numpy.array(dim_x, dim_x)
        Process noise covariance matrix. Also known as the transition
        covariance.
    F : numpy.array()
        State Transition matrix. Also known as `A` in some formulation.
    H : numpy.array(dim_z, dim_x)
        Measurement function. Also known as the observation matrix, or as `C`.
    y : numpy.array
        Residual of the update step. Read only.
    K : numpy.array(dim_x, dim_z)
        Kalman gain of the update step. Read only.
    S :  numpy.array
        System uncertainty (P projected to measurement space). Read only.
    SI :  numpy.array
        Inverse system uncertainty. Read only.
    log_likelihood : float
        log-likelihood of the last measurement. Read only.
    likelihood : float
        likelihood of last measurement. Read only.
        Computed from the log-likelihood. The log-likelihood can be very
        small,  meaning a large negative value such as -28000. Taking the
        exp() of that results in 0.0, which can break typical algorithms
        which multiply by this value, so by default we always return a
        number >= sys.float_info.min.
    mahalanobis : float
        mahalanobis distance of the innovation. Read only.
    inv : function, default numpy.linalg.inv
        If you prefer another inverse function, such as the Moore-Penrose
        pseudo inverse, set it to that instead: kf.inv = np.linalg.pinv
        This is only used to invert self.S. If you know it is diagonal, you
        might choose to set it to filterpy.common.inv_diagonal, which is
        several times faster than numpy.linalg.inv for diagonal matrices.
    alpha : float
        Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading
        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon [1]_.
    References
    ----------
    .. [1] Dan Simon. "Optimal State Estimation." John Wiley & Sons.
       p. 208-212. (2006)
    .. [2] Roger Labbe. "Kalman and Bayesian Filters in Python"
       https://github.com/rlabbe/Kalman-and-Bayesian-Filters-in-Python
    """

    def __init__(self, dim_x, dim_z, dim_u=0):
        if dim_x < 1:
            raise ValueError("dim_x must be 1 or greater")
        if dim_z < 1:
            raise ValueError("dim_z must be 1 or greater")
        if dim_u < 0:
            raise ValueError("dim_u must be 0 or greater")

        self.dim_x = dim_x
        self.dim_z = dim_z
        self.dim_u = dim_u

        self.x = zeros((dim_x, 1))  # state
        self.P = eye(dim_x)  # uncertainty covariance
        self.Q = eye(dim_x)  # process uncertainty
        self.B = None  # control transition matrix
        self.F = eye(dim_x)  # state transition matrix
        self.H = zeros((dim_z, dim_x))  # measurement function
        self.R = eye(dim_z)  # measurement uncertainty
        self._alpha_sq = 1.0  # fading memory control
        self.M = np.zeros((dim_x, dim_z))  # process-measurement cross correlation
        self.z = np.array([[None] * self.dim_z]).T

        # gain and residual are computed during the innovation step. We
        # save them so that in case you want to inspect them for various
        # purposes
        self.K = np.zeros((dim_x, dim_z))  # kalman gain
        self.y = zeros((dim_z, 1))
        self.S = np.zeros((dim_z, dim_z))  # system uncertainty
        self.SI = np.zeros((dim_z, dim_z))  # inverse system uncertainty

        # identity matrix. Do not alter this.
        self._I = np.eye(dim_x)

        # these will always be a copy of x,P after predict() is called
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

        # these will always be a copy of x,P after update() is called
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

        # Only computed only if requested via property
        self._log_likelihood = log(sys.float_info.min)
        self._likelihood = sys.float_info.min
        self._mahalanobis = None

        self.inv = np.linalg.inv

    def predict(self, u=None, B=None, F=None, Q=None):
        """
        Predict next state (prior) using the Kalman filter state propagation
        equations.
        Parameters
        ----------
        u : np.array, default 0
            Optional control vector.
        B : np.array(dim_x, dim_u), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.
        F : np.array(dim_x, dim_x), or None
            Optional state transition matrix; a value of None
            will cause the filter to use `self.F`.
        Q : np.array(dim_x, dim_x), scalar, or None
            Optional process noise matrix; a value of None will cause the
            filter to use `self.Q`.
        """

        if B is None:
            B = self.B
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        elif isscalar(Q):
            Q = eye(self.dim_x) * Q

        # x = Fx + Bu
        if B is not None and u is not None:
            self.x = dot(F, self.x) + dot(B, u)
        else:
            self.x = dot(F, self.x)

        # P = FPF' + Q
        self.P = self._alpha_sq * dot(dot(F, self.P), F.T) + Q

        # save prior
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def update(self, z, R=None, H=None):
        """
        Add a new measurement (z) to the Kalman filter.
        If z is None, nothing is computed. However, x_post and P_post are
        updated with the prior (x_prior, P_prior), and self.z is set to None.
        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
            If you pass in a value of H, z must be a column vector the
            of the correct size.
        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.
        H : np.array, or None
            Optionally provide H to override the measurement function for this
            one call, otherwise self.H will be used.
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = zeros((self.dim_z, 1))
            return

        if R is None:
            R = self.R
        elif isscalar(R):
            R = eye(self.dim_z) * R

        if H is None:
            z = reshape_z(z, self.dim_z, self.x.ndim)
            H = self.H

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - dot(H, self.x)

        # common subexpression for speed
        PHT = dot(self.P, H.T)

        # S = HPH' + R
        # project system uncertainty into measurement space
        self.S = dot(H, PHT) + R
        self.SI = self.inv(self.S)
        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = dot(PHT, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + dot(self.K, self.y)

        # P = (I-KH)P(I-KH)' + KRK'
        # This is more numerically stable
        # and works for non-optimal K vs the equation
        # P = (I-KH)P usually seen in the literature.

        I_KH = self._I - dot(self.K, H)
        self.P = dot(dot(I_KH, self.P), I_KH.T) + dot(dot(self.K, R), self.K.T)

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def predict_steadystate(self, u=0, B=None):
        """
        Predict state (prior) using the Kalman filter state propagation
        equations. Only x is updated, P is left unchanged. See
        update_steadstate() for a longer explanation of when to use this
        method.
        Parameters
        ----------
        u : np.array
            Optional control vector. If non-zero, it is multiplied by B
            to create the control input into the system.
        B : np.array(dim_x, dim_u), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.
        """

        if B is None:
            B = self.B

        # x = Fx + Bu
        if B is not None:
            self.x = dot(self.F, self.x) + dot(B, u)
        else:
            self.x = dot(self.F, self.x)

        # save prior
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def update_steadystate(self, z):
        """
        Add a new measurement (z) to the Kalman filter without recomputing
        the Kalman gain K, the state covariance P, or the system
        uncertainty S.
        You can use this for LTI systems since the Kalman gain and covariance
        converge to a fixed value. Precompute these and assign them explicitly,
        or run the Kalman filter using the normal predict()/update(0 cycle
        until they converge.
        The main advantage of this call is speed. We do significantly less
        computation, notably avoiding a costly matrix inversion.
        Use in conjunction with predict_steadystate(), otherwise P will grow
        without bound.
        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
        Examples
        --------
        >>> cv = kinematic_kf(dim=3, order=2) # 3D const velocity filter
        >>> # let filter converge on representative data, then save k and P
        >>> for i in range(100):
        >>>     cv.predict()
        >>>     cv.update([i, i, i])
        >>> saved_k = np.copy(cv.K)
        >>> saved_P = np.copy(cv.P)
        later on:
        >>> cv = kinematic_kf(dim=3, order=2) # 3D const velocity filter
        >>> cv.K = np.copy(saved_K)
        >>> cv.P = np.copy(saved_P)
        >>> for i in range(100):
        >>>     cv.predict_steadystate()
        >>>     cv.update_steadystate([i, i, i])
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = zeros((self.dim_z, 1))
            return

        z = reshape_z(z, self.dim_z, self.x.ndim)

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - dot(self.H, self.x)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + dot(self.K, self.y)

        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

    def update_correlated(self, z, R=None, H=None):
        """Add a new measurement (z) to the Kalman filter assuming that
        process noise and measurement noise are correlated as defined in
        the `self.M` matrix.
        A partial derivation can be found in [1]
        If z is None, nothing is changed.
        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
        R : np.array, scalar, or None
            Optionally provide R to override the measurement noise for this
            one call, otherwise  self.R will be used.
        H : np.array,  or None
            Optionally provide H to override the measurement function for this
            one call, otherwise  self.H will be used.
        References
        ----------
        .. [1] Bulut, Y. (2011). Applied Kalman filter theory (Doctoral dissertation, Northeastern University).
               http://people.duke.edu/~hpgavin/SystemID/References/Balut-KalmanFilter-PhD-NEU-2011.pdf
        """

        # set to None to force recompute
        self._log_likelihood = None
        self._likelihood = None
        self._mahalanobis = None

        if z is None:
            self.z = np.array([[None] * self.dim_z]).T
            self.x_post = self.x.copy()
            self.P_post = self.P.copy()
            self.y = zeros((self.dim_z, 1))
            return

        if R is None:
            R = self.R
        elif isscalar(R):
            R = eye(self.dim_z) * R

        # rename for readability and a tiny extra bit of speed
        if H is None:
            z = reshape_z(z, self.dim_z, self.x.ndim)
            H = self.H

        # handle special case: if z is in form [[z]] but x is not a column
        # vector dimensions will not match
        if self.x.ndim == 1 and shape(z) == (1, 1):
            z = z[0]

        if shape(z) == ():  # is it scalar, e.g. z=3 or z=np.array(3)
            z = np.asarray([z])

        # y = z - Hx
        # error (residual) between measurement and prediction
        self.y = z - dot(H, self.x)

        # common subexpression for speed
        PHT = dot(self.P, H.T)

        # project system uncertainty into measurement space
        self.S = dot(H, PHT) + dot(H, self.M) + dot(self.M.T, H.T) + R
        self.SI = self.inv(self.S)

        # K = PH'inv(S)
        # map system uncertainty into kalman gain
        self.K = dot(PHT + self.M, self.SI)

        # x = x + Ky
        # predict new x with residual scaled by the kalman gain
        self.x = self.x + dot(self.K, self.y)
        self.P = self.P - dot(self.K, dot(H, self.P) + self.M.T)

        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    def batch_filter(
        self,
        zs,
        Fs=None,
        Qs=None,
        Hs=None,
        Rs=None,
        Bs=None,
        us=None,
        update_first=False,
        saver=None,
    ):
        """Batch processes a sequences of measurements.
         Parameters
         ----------
         zs : list-like
             list of measurements at each time step `self.dt`. Missing
             measurements must be represented by `None`.
         Fs : None, list-like, default=None
             optional value or list of values to use for the state transition
             matrix F.
             If Fs is None then self.F is used for all epochs.
             Otherwise it must contain a list-like list of F's, one for
             each epoch.  This allows you to have varying F per epoch.
         Qs : None, np.array or list-like, default=None
             optional value or list of values to use for the process error
             covariance Q.
             If Qs is None then self.Q is used for all epochs.
             Otherwise it must contain a list-like list of Q's, one for
             each epoch.  This allows you to have varying Q per epoch.
         Hs : None, np.array or list-like, default=None
             optional list of values to use for the measurement matrix H.
             If Hs is None then self.H is used for all epochs.
             If Hs contains a single matrix, then it is used as H for all
             epochs.
             Otherwise it must contain a list-like list of H's, one for
             each epoch.  This allows you to have varying H per epoch.
         Rs : None, np.array or list-like, default=None
             optional list of values to use for the measurement error
             covariance R.
             If Rs is None then self.R is used for all epochs.
             Otherwise it must contain a list-like list of R's, one for
             each epoch.  This allows you to have varying R per epoch.
         Bs : None, np.array or list-like, default=None
             optional list of values to use for the control transition matrix B.
             If Bs is None then self.B is used for all epochs.
             Otherwise it must contain a list-like list of B's, one for
             each epoch.  This allows you to have varying B per epoch.
         us : None, np.array or list-like, default=None
             optional list of values to use for the control input vector;
             If us is None then None is used for all epochs (equivalent to 0,
             or no control input).
             Otherwise it must contain a list-like list of u's, one for
             each epoch.
        update_first : bool, optional, default=False
             controls whether the order of operations is update followed by
             predict, or predict followed by update. Default is predict->update.
         saver : filterpy.common.Saver, optional
             filterpy.common.Saver object. If provided, saver.save() will be
             called after every epoch
         Returns
         -------
         means : np.array((n,dim_x,1))
             array of the state for each time step after the update. Each entry
             is an np.array. In other words `means[k,:]` is the state at step
             `k`.
         covariance : np.array((n,dim_x,dim_x))
             array of the covariances for each time step after the update.
             In other words `covariance[k,:,:]` is the covariance at step `k`.
         means_predictions : np.array((n,dim_x,1))
             array of the state for each time step after the predictions. Each
             entry is an np.array. In other words `means[k,:]` is the state at
             step `k`.
         covariance_predictions : np.array((n,dim_x,dim_x))
             array of the covariances for each time step after the prediction.
             In other words `covariance[k,:,:]` is the covariance at step `k`.
         Examples
         --------
         .. code-block:: Python
             # this example demonstrates tracking a measurement where the time
             # between measurement varies, as stored in dts. This requires
             # that F be recomputed for each epoch. The output is then smoothed
             # with an RTS smoother.
             zs = [t + random.randn()*4 for t in range (40)]
             Fs = [np.array([[1., dt], [0, 1]] for dt in dts]
             (mu, cov, _, _) = kf.batch_filter(zs, Fs=Fs)
             (xs, Ps, Ks, Pps) = kf.rts_smoother(mu, cov, Fs=Fs)
        """

        # pylint: disable=too-many-statements
        n = np.size(zs, 0)
        if Fs is None:
            Fs = [self.F] * n
        if Qs is None:
            Qs = [self.Q] * n
        if Hs is None:
            Hs = [self.H] * n
        if Rs is None:
            Rs = [self.R] * n
        if Bs is None:
            Bs = [self.B] * n
        if us is None:
            us = [0] * n

        # mean estimates from Kalman Filter
        if self.x.ndim == 1:
            means = zeros((n, self.dim_x))
            means_p = zeros((n, self.dim_x))
        else:
            means = zeros((n, self.dim_x, 1))
            means_p = zeros((n, self.dim_x, 1))

        # state covariances from Kalman Filter
        covariances = zeros((n, self.dim_x, self.dim_x))
        covariances_p = zeros((n, self.dim_x, self.dim_x))

        if update_first:
            for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):

                self.update(z, R=R, H=H)
                means[i, :] = self.x
                covariances[i, :, :] = self.P

                self.predict(u=u, B=B, F=F, Q=Q)
                means_p[i, :] = self.x
                covariances_p[i, :, :] = self.P

                if saver is not None:
                    saver.save()
        else:
            for i, (z, F, Q, H, R, B, u) in enumerate(zip(zs, Fs, Qs, Hs, Rs, Bs, us)):

                self.predict(u=u, B=B, F=F, Q=Q)
                means_p[i, :] = self.x
                covariances_p[i, :, :] = self.P

                self.update(z, R=R, H=H)
                means[i, :] = self.x
                covariances[i, :, :] = self.P

                if saver is not None:
                    saver.save()

        return (means, covariances, means_p, covariances_p)

    def rts_smoother(self, Xs, Ps, Fs=None, Qs=None, inv=np.linalg.inv):
        """
        Runs the Rauch-Tung-Striebel Kalman smoother on a set of
        means and covariances computed by a Kalman filter. The usual input
        would come from the output of `KalmanFilter.batch_filter()`.
        Parameters
        ----------
        Xs : numpy.array
           array of the means (state variable x) of the output of a Kalman
           filter.
        Ps : numpy.array
            array of the covariances of the output of a kalman filter.
        Fs : list-like collection of numpy.array, optional
            State transition matrix of the Kalman filter at each time step.
            Optional, if not provided the filter's self.F will be used
        Qs : list-like collection of numpy.array, optional
            Process noise of the Kalman filter at each time step. Optional,
            if not provided the filter's self.Q will be used
        inv : function, default numpy.linalg.inv
            If you prefer another inverse function, such as the Moore-Penrose
            pseudo inverse, set it to that instead: kf.inv = np.linalg.pinv
        Returns
        -------
        x : numpy.ndarray
           smoothed means
        P : numpy.ndarray
           smoothed state covariances
        K : numpy.ndarray
            smoother gain at each step
        Pp : numpy.ndarray
           Predicted state covariances
        Examples
        --------
        .. code-block:: Python
            zs = [t + random.randn()*4 for t in range (40)]
            (mu, cov, _, _) = kalman.batch_filter(zs)
            (x, P, K, Pp) = rts_smoother(mu, cov, kf.F, kf.Q)
        """

        if len(Xs) != len(Ps):
            raise ValueError("length of Xs and Ps must be the same")

        n = Xs.shape[0]
        dim_x = Xs.shape[1]

        if Fs is None:
            Fs = [self.F] * n
        if Qs is None:
            Qs = [self.Q] * n

        # smoother gain
        K = zeros((n, dim_x, dim_x))

        x, P, Pp = Xs.copy(), Ps.copy(), Ps.copy()
        for k in range(n - 2, -1, -1):
            Pp[k] = dot(dot(Fs[k + 1], P[k]), Fs[k + 1].T) + Qs[k + 1]

            # pylint: disable=bad-whitespace
            K[k] = dot(dot(P[k], Fs[k + 1].T), inv(Pp[k]))
            x[k] += dot(K[k], x[k + 1] - dot(Fs[k + 1], x[k]))
            P[k] += dot(dot(K[k], P[k + 1] - Pp[k]), K[k].T)

        return (x, P, K, Pp)

    def get_prediction(self, u=None, B=None, F=None, Q=None):
        """
        Predict next state (prior) using the Kalman filter state propagation
        equations and returns it without modifying the object.
        Parameters
        ----------
        u : np.array, default 0
            Optional control vector.
        B : np.array(dim_x, dim_u), or None
            Optional control transition matrix; a value of None
            will cause the filter to use `self.B`.
        F : np.array(dim_x, dim_x), or None
            Optional state transition matrix; a value of None
            will cause the filter to use `self.F`.
        Q : np.array(dim_x, dim_x), scalar, or None
            Optional process noise matrix; a value of None will cause the
            filter to use `self.Q`.
        Returns
        -------
        (x, P) : tuple
            State vector and covariance array of the prediction.
        """

        if B is None:
            B = self.B
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        elif isscalar(Q):
            Q = eye(self.dim_x) * Q

        # x = Fx + Bu
        if B is not None and u is not None:
            x = dot(F, self.x) + dot(B, u)
        else:
            x = dot(F, self.x)

        # P = FPF' + Q
        P = self._alpha_sq * dot(dot(F, self.P), F.T) + Q

        return x, P

    def get_update(self, z=None):
        """
        Computes the new estimate based on measurement `z` and returns it
        without altering the state of the filter.
        Parameters
        ----------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
        Returns
        -------
        (x, P) : tuple
            State vector and covariance array of the update.
        """

        if z is None:
            return self.x, self.P
        z = reshape_z(z, self.dim_z, self.x.ndim)

        R = self.R
        H = self.H
        P = self.P
        x = self.x

        # error (residual) between measurement and prediction
        y = z - dot(H, x)

        # common subexpression for speed
        PHT = dot(P, H.T)

        # project system uncertainty into measurement space
        S = dot(H, PHT) + R

        # map system uncertainty into kalman gain
        K = dot(PHT, self.inv(S))

        # predict new x with residual scaled by the kalman gain
        x = x + dot(K, y)

        # P = (I-KH)P(I-KH)' + KRK'
        I_KH = self._I - dot(K, H)
        P = dot(dot(I_KH, P), I_KH.T) + dot(dot(K, R), K.T)

        return x, P

    def residual_of(self, z):
        """Returns the residual for the given measurement (z).

        Does not alter the state of the filter.
        """
        z = reshape_z(z, self.dim_z, self.x.ndim)
        return z - dot(self.H, self.x_prior)

    def measurement_of_state(self, x):
        """
        Helper function that converts a state into a measurement.
        Parameters
        ----------
        x : np.array
            kalman state vector
        Returns
        -------
        z : (dim_z, 1): array_like
            measurement for this update. z can be a scalar if dim_z is 1,
            otherwise it must be convertible to a column vector.
        """

        return dot(self.H, x)

    @property
    def log_likelihood(self):
        """log-likelihood of the last measurement."""
        if self._log_likelihood is None:
            self._log_likelihood = logpdf(x=self.y, cov=self.S)
        return self._log_likelihood

    @property
    def likelihood(self):
        """Computed from the log-likelihood.

        The log-likelihood can be very small,  meaning a large negative
        value such as -28000. Taking the exp() of that results in 0.0,
        which can break typical algorithms which multiply by this value,
        so by default we always return a number >= sys.float_info.min.
        """
        if self._likelihood is None:
            self._likelihood = exp(self.log_likelihood)
            if self._likelihood == 0:
                self._likelihood = sys.float_info.min
        return self._likelihood

    @property
    def mahalanobis(self):
        """
        Mahalanobis distance of measurement. E.g. 3 means measurement
        was 3 standard deviations away from the predicted value.
        Returns
        -------
        mahalanobis : float
        """
        if self._mahalanobis is None:
            self._mahalanobis = sqrt(float(dot(dot(self.y.T, self.SI), self.y)))
        return self._mahalanobis

    @property
    def alpha(self):
        """Fading memory setting. 1.0 gives the normal Kalman filter, and
        values slightly larger than 1.0 (such as 1.02) give a fading.

        memory effect - previous measurements have less influence on the
        filter's estimates. This formulation of the Fading memory filter
        (there are many) is due to Dan Simon [1]_.
        """
        return self._alpha_sq ** 0.5

    def log_likelihood_of(self, z):
        """log likelihood of the measurement `z`.

        This should only be called after a call to update(). Calling
        after predict() will yield an incorrect result.
        """

        if z is None:
            return log(sys.float_info.min)
        return logpdf(z, dot(self.H, self.x), self.S)

    @alpha.setter
    def alpha(self, value):
        if not np.isscalar(value) or value < 1:
            raise ValueError("alpha must be a float greater than 1")

        self._alpha_sq = value ** 2

    def __repr__(self):
        return "\n".join(
            [
                "KalmanFilter object",
                pretty_str("dim_x", self.dim_x),
                pretty_str("dim_z", self.dim_z),
                pretty_str("dim_u", self.dim_u),
                pretty_str("x", self.x),
                pretty_str("P", self.P),
                pretty_str("x_prior", self.x_prior),
                pretty_str("P_prior", self.P_prior),
                pretty_str("x_post", self.x_post),
                pretty_str("P_post", self.P_post),
                pretty_str("F", self.F),
                pretty_str("Q", self.Q),
                pretty_str("R", self.R),
                pretty_str("H", self.H),
                pretty_str("K", self.K),
                pretty_str("y", self.y),
                pretty_str("S", self.S),
                pretty_str("SI", self.SI),
                pretty_str("M", self.M),
                pretty_str("B", self.B),
                pretty_str("z", self.z),
                pretty_str("log-likelihood", self.log_likelihood),
                pretty_str("likelihood", self.likelihood),
                pretty_str("mahalanobis", self.mahalanobis),
                pretty_str("alpha", self.alpha),
                pretty_str("inv", self.inv),
            ]
        )

    def test_matrix_dimensions(self, z=None, H=None, R=None, F=None, Q=None):
        """Performs a series of asserts to check that the size of everything is
        what it should be.

        This can help you debug problems in your design. If you pass in
        H, R, F, Q those will be used instead of this object's value for
        those matrices. Testing `z` (the measurement) is problamatic. x
        is a vector, and can be implemented as either a 1D array or as a
        nx1 column vector. Thus Hx can be of different shapes. Then, if
        Hx is a single value, it can be either a 1D array or 2D vector.
        If either is true, z can reasonably be a scalar (either '3' or
        np.array('3') are scalars under this definition), a 1D, 1
        element array, or a 2D, 1 element array. You are allowed to pass
        in any combination that works.
        """

        if H is None:
            H = self.H
        if R is None:
            R = self.R
        if F is None:
            F = self.F
        if Q is None:
            Q = self.Q
        x = self.x
        P = self.P

        assert (
            x.ndim == 1 or x.ndim == 2
        ), "x must have one or two dimensions, but has {}".format(x.ndim)

        if x.ndim == 1:
            assert (
                x.shape[0] == self.dim_x
            ), "Shape of x must be ({},{}), but is {}".format(self.dim_x, 1, x.shape)
        else:
            assert x.shape == (
                self.dim_x,
                1,
            ), "Shape of x must be ({},{}), but is {}".format(self.dim_x, 1, x.shape)

        assert P.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of P must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, P.shape
        )

        assert Q.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of Q must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, P.shape
        )

        assert F.shape == (
            self.dim_x,
            self.dim_x,
        ), "Shape of F must be ({},{}), but is {}".format(
            self.dim_x, self.dim_x, F.shape
        )

        assert np.ndim(H) == 2, "Shape of H must be (dim_z, {}), but is {}".format(
            P.shape[0], shape(H)
        )

        assert (
            H.shape[1] == P.shape[0]
        ), "Shape of H must be (dim_z, {}), but is {}".format(P.shape[0], H.shape)

        # shape of R must be the same as HPH'
        hph_shape = (H.shape[0], H.shape[0])
        r_shape = shape(R)

        if H.shape[0] == 1:
            # r can be scalar, 1D, or 2D in this case
            assert r_shape in [
                (),
                (1,),
                (1, 1),
            ], "R must be scalar or one element array, but is shaped {}".format(r_shape)
        else:
            assert r_shape == hph_shape, "shape of R should be {} but it is {}".format(
                hph_shape, r_shape
            )

        if z is not None:
            z_shape = shape(z)
        else:
            z_shape = (self.dim_z, 1)

        # H@x must have shape of z
        Hx = dot(H, x)

        if z_shape == ():  # scalar or np.array(scalar)
            assert Hx.ndim == 1 or shape(Hx) == (
                1,
                1,
            ), "shape of z should be {}, not {} for the given H".format(
                shape(Hx), z_shape
            )

        elif shape(Hx) == (1,):
            assert z_shape[0] == 1, "Shape of z must be {} for the given H".format(
                shape(Hx)
            )

        else:
            assert z_shape == shape(Hx) or (
                len(z_shape) == 1 and shape(Hx) == (z_shape[0], 1)
            ), "shape of z should be {}, not {} for the given H".format(
                shape(Hx), z_shape
            )

        if np.ndim(Hx) > 1 and shape(Hx) != (1, 1):
            assert (
                shape(Hx) == z_shape
            ), "shape of z should be {} for the given H, but it is {}".format(
                shape(Hx), z_shape
            )



import json
from inspect import getfullargspec, ismethod
from time import time

from flask import Blueprint


def log_unhandled_message(message_type, message_data):
    if message_data is None:
        print(f'Unhandled message "{message_type}"')
        return

    pretty_message_data = json.dumps(message_data, indent=4, sort_keys=True)
    print(f'Unhandled message "{message_type}": {pretty_message_data}')


class MessagingBlueprint(Blueprint):
    def __init__(self, message_handlers={}, **kwargs):
        Blueprint.__init__(
            self,
            "messaging",
            __name__,
            static_folder="messaging",
            template_folder="templates",
            **kwargs,
        )

        def handle_message(message, send):
            parsed_message = json.loads(message)

            message_type = parsed_message.get("type", "")
            message_data = parsed_message.get("data")

            handler = message_handlers.get(message_type)

            if handler is None:
                log_unhandled_message(message_type, message_data)
                return

            spec = getfullargspec(handler)
            if len(spec.args) == (3 if ismethod(handler) else 2) or spec.varargs:
                handler(message_data, send)
                return

            if len(spec.args) == (2 if ismethod(handler) else 1):
                handler(message_data)
                return

            handler()

        self.sockets = {}
        self.socket_blueprint = Blueprint("messaging_socket", __name__)

        @self.socket_blueprint.route("/messaging")
        def pubsub(ws):
            id = time()
            self.sockets[id] = ws

            def send(response_message):
                ws.send(json.dumps(response_message))

            while not ws.closed:
                message = ws.receive()
                if message:
                    handle_message(message, send)

            del self.sockets[id]

    def register(self, app, options, *args):
        Blueprint.register(self, app, options, *args)

        sockets = options.get("sockets")
        if sockets is None:
            raise Exception("Unable to register MessagingBlueprint without sockets")

        sockets.register_blueprint(self.socket_blueprint)

    def broadcast(self, message):
        for socket in self.sockets.values():
            socket.send(json.dumps(message))



from atexit import register
from threading import Thread, current_thread, main_thread
from time import sleep

from pitop.core import ImageFunctions

from .assistant import MiniscreenAssistant
from .core import FPS_Regulator, MiniscreenLockFileMonitor, OledDeviceController


class OLED:
    """Provides access to the miniscreen display on the pi-top [4], and exposes
    methods for simple rendering of text or images to the screen."""

    def __init__(self):
        self.__controller = OledDeviceController(self._redraw_last_image)
        self.lock_file_monitor = MiniscreenLockFileMonitor(self.__controller.lock.path)

        self.assistant = MiniscreenAssistant(self.mode, self.size)

        self.image = self.assistant.empty_image
        self._image = self.assistant.empty_image

        self.__fps_regulator = FPS_Regulator()

        self.__visible = False
        self.__auto_play_thread = None

        self.reset()

        register(self.__cleanup)

    def prepare_image(self, image_to_prepare):
        """Formats the given image into one that can be used directly by the
        OLED.

        :type image_to_prepare: :class:`PIL.Image.Image`
        :param image_to_prepare: Image to be formatted.
        :rtype: :class:`PIL.Image.Image`
        """
        # TODO: deprecate this function in favour of user handling this directly
        return self.assistant.process_image(image_to_prepare)

    def should_redisplay(self, image_to_display=None):
        """Determines if the miniscreen display needs to be refreshed, based on
        the provided image. If no image is provided, the content of the
        display's deprecated internal canvas property will be used.

        :type image_to_display: :class:`PIL.Image.Image` or None
        :param image_to_display: Image to be displayed.
        :rtype: bool
        """
        # Use canvas image if no image is offered
        if image_to_display is None:
            image_to_display = self._image

        return self.image is None or self.assistant.images_match(
            self.image, image_to_display
        )

    @property
    def spi_bus(self):
        """Gets the SPI bus used by the miniscreen display to receive data as
        an integer. Setting this property will modify the SPI bus that the OLED
        uses. You might notice a flicker in the screen.

        :param int bus: Number of the SPI bus for the OLED to use. Accepted values are `0` or `1`.
        """
        return self.__controller.spi_bus

    @spi_bus.setter
    def spi_bus(self, bus):
        assert bus in (0, 1)
        self.__controller.spi_bus = bus

    @property
    def device(self):
        """Gets the miniscreen display device instance.

        :rtype: :class:`pitop.miniscreen.oled.core.contrib.luma.oled.device.sh1106`
        """
        return self.__controller.get_device()

    @property
    def size(self):
        """Gets the size of the miniscreen display as a (width, height) tuple.

        :rtype: tuple
        """
        return self.device.size

    @property
    def width(self):
        """Gets the width of the miniscreen display.

        :rtype: int
        """
        return self.size[0]

    @property
    def height(self):
        """Gets the height of the miniscreen display.

        :rtype: int
        """
        return self.size[1]

    @property
    def mode(self):
        return self.device.mode

    @property
    def is_active(self):
        """Determine if the current OLED instance is in control of the OLED
        hardware.

        :return: whether the OLED instance is in control of the OLED hardware.
        :rtype: bool
        """
        return self.__controller.device_is_active()

    @property
    def visible(self):
        """Gets whether the device is currently in low power state.

        :return: whether the screen is in low power mode
        :rtype: bool
        """
        return not self.__visible

    def set_control_to_pi(self):
        """Signals the pi-top hub to give control of the miniscreen display to
        the Raspberry Pi."""
        self.__controller.set_control_to_pi()

    def set_control_to_hub(self):
        """Signals the pi-top hub to take control of the miniscreen display."""
        self.__controller.set_control_to_hub()

    def set_max_fps(self, max_fps):
        """Set the maximum frames per second that the miniscreen display can
        display. This method can be useful to control or limit the speed of
        animations.

        This works by blocking on the OLED's display methods if called before
        the amount of time that a frame should last is not exceeded.

        :param int max_fps: The maximum frames that can be rendered per second
        """
        self.__fps_regulator.set_max_fps(max_fps)

    def show(self):
        """The miniscreen display comes out of low power mode showing the
        previous image shown before hide() was called (so long as display() has
        not been called)"""
        self.device.show()
        self.__visible = True

    def hide(self):
        """The miniscreen display is put into low power mode.

        The previously shown image will re-appear when show() is given,
        even if the internal frame buffer has been changed (so long as
        display() has not been called).
        """
        self.device.hide()
        self.__visible = False

    def contrast(self, new_contrast_value):
        """Sets the contrast value of the miniscreen display to the provided
        value.

        :param int new_contrast_value: contrast value to set, between 0 and 255.
        """
        assert new_contrast_value in range(0, 256)

        self.device.contrast(new_contrast_value)

    def wake(self):
        """The miniscreen display is set to high contrast mode, without
        modifying the content of the screen."""
        self.contrast(255)

    def sleep(self):
        """The miniscreen display in set to low contrast mode, without
        modifying the content of the screen."""
        self.contrast(0)

    def clear(self):
        """Clears any content displayed in the miniscreen display."""
        self.assistant.clear(self._image)
        self.__display(self._image, force=True)

    # TODO: evaluate dropping this 'redraw last image' function at v1.0.0
    #
    # this is only necessary to support users with SPI0 on device
    # with older SDK version that only supported SPI1
    def _redraw_last_image(self):
        self.__display(self.image, force=True)

    def refresh(self):
        self.set_control_to_pi()
        self.__controller.reset_device()
        self._redraw_last_image()

    def reset(self, force=True):
        """Gives the caller access to the miniscreen display (i.e. in the case
        the system is currently rendering information to the screen) and clears
        the screen."""
        self.clear()
        self.refresh()

        self.wake()

        if not self.visible:
            self.show()

    def display_image_file(self, file_path_or_url, xy=None, invert=False):
        """Render a static image to the screen from a file or URL at a given
        position.

        The display's positional properties (e.g. `top_left`, `top_right`) can be used to assist with
        specifying the `xy` position parameter.

        :param str file_path_or_url: A file path or URL to the image
        :param tuple xy: The position on the screen to render the image. If not
            provided or passed as `None` the image will be drawn in the top-left of
            the screen.
        :param bool invert: Set to True to flip the on/off state of each pixel in the image
        """
        self.display_image(
            ImageFunctions.get_pil_image_from_path(file_path_or_url),
            xy=xy,
            invert=invert,
        )

    # TODO: add 'size' parameter
    # TODO: add 'fill', 'stretch', 'crop', etc. to OLED images - currently, they only stretch by default
    # TODO: handle 'xy'
    def display_image(self, image, xy=None, invert=False):
        """Render a static image to the screen from a file or URL at a given
        position.

        The image should be provided as a PIL Image object.

        :param Image image: A PIL Image object to be rendered
        :param tuple xy: The position on the screen to render the image. If not
            provided or passed as `None` the image will be drawn in the top-left of
            the screen.
        :param bool invert: Set to True to flip the on/off state of each pixel in the image
        """
        self.__display(
            self.assistant.process_image(ImageFunctions.convert(image, format="PIL")),
            invert=invert,
        )

    def display_text(
        self,
        text,
        xy=None,
        font_size=None,
        font=None,
        invert=False,
        align=None,
        anchor=None,
    ):
        """Renders a single line of text to the screen at a given position and
        size.

        The display's positional properties (e.g. `top_left`, `top_right`) can be used to assist with
        specifying the `xy` position parameter.

        :param string text: The text to render
        :param tuple xy: The position on the screen to render the image. If not
            provided or passed as `None` the image will be drawn in the top-left of
            the screen.
        :param int font_size: The font size in pixels. If not provided or passed as
            `None`, the default font size will be used
        :param string font: A filename or path of a TrueType or OpenType font.
            If not provided or passed as `None`, the default font will be used
        :param bool invert: Set to True to flip the on/off state of each pixel in the image
        :param str align: PIL ImageDraw alignment to use
        :param str anchor: PIL ImageDraw text anchor to use
        """
        image = self.assistant.empty_image
        self.assistant.render_text(
            image,
            text,
            xy,
            font_size,
            font,
            align,
            anchor,
        )
        self.display_image(image, invert=invert)

    def display_multiline_text(
        self,
        text,
        xy=None,
        font_size=None,
        font=None,
        invert=False,
        anchor=None,
        align=None,
    ):
        """Renders multi-lined text to the screen at a given position and size.
        Text that is too long for the screen will automatically wrap to the
        next line.

        The display's positional properties (e.g. `top_left`, `top_right`) can be used to assist with
        specifying the `xy` position parameter.

        :param string text: The text to render
        :param tuple xy: The position on the screen to render the image. If not
            provided or passed as `None` the image will be drawn in the top-left of
            the screen.
        :param int font_size: The font size in pixels. If not provided or passed as
            `None`, the default font size will be used
        :param string font: A filename or path of a TrueType or OpenType font.
            If not provided or passed as `None`, the default font will be used
        :param bool invert: Set to True to flip the on/off state of each pixel in the image
        :param str align: PIL ImageDraw alignment to use
        :param str anchor: PIL ImageDraw text anchor to use
        """
        image = self.assistant.empty_image
        self.assistant.render_multiline_text(
            image,
            text,
            xy,
            font_size,
            font,
            align,
            anchor,
        )
        self.display_image(image, invert=invert)

    def __display(self, image_to_display, force=False, invert=False):
        self.stop_animated_image()

        if invert:
            image_to_display = self.assistant.invert(image_to_display)

        self.__fps_regulator.stop_timer()

        if force or self.should_redisplay(image_to_display):
            self.device.display(image_to_display)

        self.__fps_regulator.start_timer()
        self.image = image_to_display.copy()

    def play_animated_image_file(self, file_path_or_url, background=False, loop=False):
        """Render an animated image to the screen from a file or URL.

        :param str file_path_or_url: A file path or URL to the image
        :param bool background: Set whether the image should be in a background thread
            or in the main thread.
        :param bool loop: Set whether the image animation should start again when it
            has finished
        """
        image = ImageFunctions.get_pil_image_from_path(file_path_or_url)
        self.play_animated_image(image, background, loop)

    def play_animated_image(self, image, background=False, loop=False):
        """Render an animation or a image to the screen.

        Use stop_animated_image() to end a background animation

        :param Image image: A PIL Image object to be rendered
        :param bool background: Set whether the image should be in a background thread
            or in the main thread.
        :param bool loop: Set whether the image animation should start again when it
            has finished
        """
        self.stop_animated_image()
        self.__kill_thread = False
        if background is True:
            self.__auto_play_thread = Thread(
                target=self.__auto_play, args=(image, loop)
            )
            self.__auto_play_thread.start()
        else:
            self.__auto_play(image, loop)

    def stop_animated_image(self):
        """Stop background animation started using `start()`, if currently
        running."""
        if current_thread() is not main_thread():
            # thread that runs an animation in the background can't "join" itself
            return

        if self.__auto_play_thread is not None and self.__auto_play_thread.is_alive():
            self.__kill_thread = True
            self.__auto_play_thread.join()

    ##################################################
    # Position/dimension methods
    ##################################################
    @property
    def bounding_box(self):
        """Gets the bounding box of the miniscreen display.

        :return: The device's bounding box as an (top-left x, top-left y, bottom-right x, bottom-right y) tuple.
        :rtype: tuple
        """
        return self.device.bounding_box

    @property
    def center(self):
        """Gets the center of the miniscreen display.

        :return: The coordinates of the center of the display's bounding box as an (x,y) tuple.
        :rtype: tuple
        """
        return (self.width / 2, self.height / 2)

    @property
    def top_left(self):
        """Gets the top left corner of the miniscreen display.

        :return: The coordinates of the center of the display's bounding box as an (x,y) tuple.
        :rtype: tuple
        """
        return (self.bounding_box[0], self.bounding_box[1])

    @property
    def top_right(self):
        """Gets the top-right corner of the miniscreen display.

        :return: The coordinates of the top right of the display's bounding box as an (x,y) tuple.
        :rtype: tuple
        """
        return (self.bounding_box[2], self.bounding_box[1])

    @property
    def bottom_left(self):
        """Gets the bottom-left corner of the miniscreen display.

        :return: The coordinates of the bottom left of the display's bounding box as an (x,y) tuple.
        :rtype: tuple
        """
        return (self.bounding_box[0], self.bounding_box[3])

    @property
    def bottom_right(self):
        """Gets the bottom-right corner of the miniscreen display.

        :return: The coordinates of the bottom right of the display's bounding box as an (x,y) tuple.
        :rtype: tuple
        """
        return (self.bounding_box[2], self.bounding_box[3])

    #######################
    # Deprecation support #
    #######################
    def display(self, force=False):
        """Displays what is on the current canvas to the screen as a single
        frame.

        .. warning::
           This method is deprecated and will be deleted on the next major release of the SDK.

        This method does not need to be called when using the other `draw`
        functions in this class, but is used when the caller wants to use
        the *canvas* object to draw composite objects and then render them
        to screen in a single frame.
        """
        print(
            "'display()' is now deprecated. You will need to handle your own images in future."
        )
        self.__display(self._image, force=force)

    def draw(self):
        """
        .. warning::
           This method is deprecated in favor of :func:`display_image` and
           :func:`display_text`, and will be deleted on the next major release of the SDK.
        """
        print("'draw()' is now deprecated. Using 'display()'...")
        self.display()

    def draw_image_file(self, file_path_or_url, xy=None):
        """
        .. warning::
           This method is deprecated in favor of :func:`display_image_file`, and will be deleted on the next major release of the SDK.
        """
        print("draw_image_file is now deprecated. Using display_image_file...")
        self.display_image_file(file_path_or_url, xy)

    def draw_image(self, image, xy=None):
        """
        .. warning::
           This method is deprecated in favor of :func:`display_image`, and will be deleted on the next major release of the SDK.
        """
        print("draw_image is now deprecated. Using display_image...")
        self.display_image(image, xy)

    def draw_text(self, text, xy=None, font_size=None):
        """
        .. warning::
           This method is deprecated in favor of :func:`display_text`, and will be deleted on the next major release of the SDK.
        """
        print("draw_text is now deprecated. Using display_text...")
        self.display_text(text, xy, font_size)

    def draw_multiline_text(self, text, xy=None, font_size=None):
        """
        .. warning::
           This method is deprecated in favor of :func:`display_multiline_text`, and will be deleted on the next major release of the SDK.
        """
        print("draw_multiline_text is now deprecated. Using display_multiline_text...")
        self.display_multiline_text(text, xy, font_size)

    ####################
    # Internal support #
    ####################
    def __auto_play(self, image, loop=False):
        while True:
            for frame in self.assistant.get_frame_iterator(image):

                if self.__kill_thread:
                    break

                self.display_image(frame)
                # Wait for animated image's frame length
                sleep(float(frame.info["duration"] / 1000))  # ms to s

            if self.__kill_thread or not loop:
                self.reset()
                break

    @property
    def when_user_controlled(self):
        """Function to call when user takes control of the miniscreen.

        This is used by pt-miniscreen to update its 'user-controlled'
        application state.
        """
        return self.lock_file_monitor.when_user_starts_using_oled

    @when_user_controlled.setter
    def when_user_controlled(self, callback):
        """Setter for function to call when user takes control of the
        miniscreen.

        This is used by pt-miniscreen to update its 'user-controlled'
        application state.
        """
        if not callable(callback):
            raise ValueError("Callback must be callable")

        self.lock_file_monitor.when_user_starts_using_oled = callback
        # Lockfile thread needs to be restarted to get updated callback reference
        self.lock_file_monitor.start()

    @property
    def when_system_controlled(self):
        """Function to call when user gives back control of the miniscreen to
        the system.

        This is used by pt-miniscreen to update its 'user-controlled'
        application state.
        """
        return self.lock_file_monitor.when_user_stops_using_oled

    @when_system_controlled.setter
    def when_system_controlled(self, callback):
        """Setter for function to call when user gives back control of the
        miniscreen to the system.

        This is used by pt-miniscreen to update its 'user-controlled'
        application state.
        """
        if not callable(callback):
            raise ValueError("Callback must be callable")

        self.lock_file_monitor.when_user_stops_using_oled = callback
        # Lockfile thread needs to be restarted to get updated callback reference
        self.lock_file_monitor.start()

    @property
    def _when_user_starts_using_oled(self):
        """Deprecated function."""
        return self.when_user_controlled

    @_when_user_starts_using_oled.setter
    def _when_user_starts_using_oled(self, callback):
        """Deprecated function."""
        self.when_user_controlled = callback

    @property
    def _when_user_stops_using_oled(self):
        """Deprecated function."""
        return self.when_system_controlled

    @_when_user_stops_using_oled.setter
    def _when_user_stops_using_oled(self, callback):
        """Deprecated function."""
        self.when_system_controlled = callback

    def __cleanup(self):
        self.stop_animated_image()
        self.lock_file_monitor.stop()



import argparse
from os import get_terminal_size
from sys import stderr

from .cli_base import CliBaseClass


class StdoutFormat:
    BOLD = "\033[1m"
    ENDC = "\033[0m"
    GREEN = "\033[92m"


class DeviceCLI(CliBaseClass):
    parser_help = "Get information about device and attached pi-top hardware"
    cli_name = "devices"

    def __init__(self, args) -> None:
        self.args = args

    def run(self) -> int:
        def print_header(section_name):
            print(
                f"{StdoutFormat.BOLD}{section_name}{StdoutFormat.ENDC} {'='*(get_terminal_size().columns - len(section_name) - 2)}"
            )

        def print_peripheral_line(data):
            if data.get("connected") is False and self.args.quiet:
                return

            if self.args.devices_subcommand is None or not self.args.quiet:
                print(
                    f"[ {StdoutFormat.GREEN}{'✓' if data.get('connected') else ' '}{StdoutFormat.ENDC} ]",
                    end=" ",
                )
            print(f"{data.get('name')}", end=" ")
            if not self.args.name_only and data.get("fw_version"):
                print(f"(v{data.get('fw_version')})", end="")
            print("")

        def print_hub_line(data):
            print(f"{data.get('name')}", end="")
            if not self.args.name_only and data.get("fw_version"):
                print(f" (v{data.get('fw_version')})", end="")
            print("")

        # Get host device from pi-topd
        try:
            from pitop.system import device_info

            device = device_info()
            if self.args.devices_subcommand in ("hub", None):
                if self.args.devices_subcommand is None:
                    print_header("HUB")
                print_hub_line(device)
        except Exception as e:
            print(
                f"Error on pitop-devices.run: Unable to get device type from pi-topd: {e}",
                file=stderr,
            )
            return 1

        if self.args.devices_subcommand in ("peripherals", None):
            if self.args.devices_subcommand is None:
                print_header("PERIPHERALS")

            try:
                # Get list of all pi-top peripherals
                from pitop.system import pitop_peripherals

                for periph in pitop_peripherals():
                    print_peripheral_line(periph)
            except Exception as e:
                print(
                    f"Error on pitop-devices.run: Unable to get connected peripherals from pi-topd: {e}",
                    file=stderr,
                )
        return 0

    @classmethod
    def add_parser_arguments(cls, parser) -> None:
        def add_common_arguments(parser):
            parser.add_argument(
                "--quiet",
                "-q",
                help="Display only the connected devices",
                action="store_true",
            )
            parser.add_argument(
                "--name-only",
                "-n",
                help="Display only the name of the devices, without further information",
                action="store_true",
            )

        # to use arguments with "devices" directly
        add_common_arguments(parser)

        # manage arguments common to subparser options (hub & peripherals)
        parent_parser = argparse.ArgumentParser(add_help=False)
        add_common_arguments(parent_parser)

        subparser = parser.add_subparsers(
            title="pi-top devices utility",
            description="Get information about pi-top attached devices",
            dest="devices_subcommand",
        )
        # "pitop devices hub" subcommand
        subparser.add_parser(
            "hub",
            help="Get the name of the active pi-top device",
            parents=[parent_parser],
        )
        # "pitop devices peripherals" subcommand
        subparser.add_parser(
            "peripherals",
            help="Get information about attached pi-top peripherals",
            parents=[parent_parser],
        )


def main():
    from .deprecated_cli_runner import run

    run(DeviceCLI)


def host():
    from .deprecated_cli_runner import run_with_args

    args = {"devices_subcommand": "hub", "name_only": True}
    run_with_args(
        DeviceCLI,
        old_command="pt-host",
        new_command="pi-top devices hub",
        args_dict=args,
    )


if __name__ == "__main__":
    main()



# coding=utf-8
# pystray
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

__author__ = u"Moses Palmér"
__version__ = (1, 4, 2)



from .recreatable import Recreatable
from .stateful import Stateful


class Componentable(Stateful, Recreatable):
    """Represents an object that attach components to itself, and that can
    store its internal configuration to recreate itself in the future."""

    def __init__(self, children=[], config_dict=None):
        if config_dict is None:
            config_dict = dict()

        Stateful.__init__(self, children)
        Recreatable.__init__(self, config_dict)

    @classmethod
    def from_config(cls, config_dict):
        """Creates an instance of a :class:`Componentable` using the provided
        dictionary.

        If a component fails to be recreated, the main object will still
        be created, but without it.
        """
        components_config = config_dict.pop("components", dict())
        # Create main object
        main_obj = super().from_config(config_dict)

        # Add components
        for name, config in components_config.items():
            try:
                component = super().from_config(config)
                component.name = name
                main_obj.add_component(component)
            except Exception as e:
                print(f"Error recreating {name}: {e}")

        return main_obj

    def add_component(self, component, name=""):
        """Attaches a component to the current instance as an attribute. It can
        be accessed using the component name or the name provided in the
        arguments.

        Note that even though non native pi-top components can be
        added to a instance, only components that inherit from
        :class:`Recreatable` will be displayed in the :prop:`config`.
        """
        is_recreatable = isinstance(component, Recreatable)

        if not is_recreatable and not name:
            raise AttributeError("A name must be provided to add this component.")
        if name and not isinstance(name, str):
            raise AttributeError("Name must be a string")

        if name and is_recreatable:
            component.name = name

        component_name = component.name if is_recreatable else name
        if component_name in self.children:
            raise AttributeError(
                f"There's already a component with the name '{component_name}' registered."
            )

        self.children.append(component_name)
        setattr(self, component_name, component)

    @property
    def config(self):
        """Returns the component configuration as a dictionary."""
        cfg = super().config
        cfg["components"] = dict()
        for child_name in self.children:
            if hasattr(self, child_name):
                child_obj = getattr(self, child_name)
                if isinstance(child_obj, Recreatable):
                    cfg["components"][child_name] = child_obj.config
        return cfg



import os

from pitop.core import ImageFunctions


class FsImage:
    def __init__(self, path: str):
        self.path = path
        self.data = ImageFunctions.get_pil_image_from_path(self.path)
        if self.data is None:
            raise AttributeError(f"Couldn't load image {path}")


class FileSystemCamera:
    """Class that mocks the behaviour of a camera, by using a set of pictures
    from the given :data:`path`

    :param str path: path where images to be used by the camera are located in the system
    :param str file_starts_with: filter images from the given path to use only the ones that start with the given string
    :param str file_ends_with: filter images from the given path to use only the ones that end with the given string
    """

    def __init__(
        self, path: str = "/tmp", file_starts_with: str = "", file_ends_with: str = ""
    ):
        if not os.path.isdir(path):
            raise FileExistsError(f"Provided path {path} doesn't exist")

        self.path = path
        self.__images_arr = []
        self.__current_index = 0
        self.__current_image = None

        self.__file_starts_with = file_starts_with
        self.__file_ends_with = file_ends_with

        self.__load_images()
        self.__load_frame(self.__current_index)

    def __load_images(self, starts_with: str = "", ends_with: str = ""):
        self.__images_arr = []

        with os.scandir(self.path) as it:
            for file_path in it:
                if not (
                    file_path.name.startswith(self.__file_starts_with)
                    or file_path.name.endswith(self.__file_ends_with)
                ):
                    continue

                try:
                    image_obj = FsImage(file_path.path)
                    self.__images_arr.append(image_obj)
                except Exception:
                    continue

    def reset(self):
        self.__current_index = 0
        self.__load_frame(self.__current_index)

    def __advance(self):
        index = min(self.__current_index + 1, len(self.__images_arr) - 1)
        if index == self.__current_index:
            return

        self.__current_index = index
        self.__load_frame(self.__current_index)

    def __load_frame(self, index: int):
        if 0 <= index < len(self.__images_arr):
            self.__current_image = self.__images_arr[index]

    def current_frame_source(self):
        return self.__current_image.path

    def is_opened(self):
        return self.__current_image is not None

    def get_frame(self):
        self.__advance()
        return self.__current_image.data



from pitop.common.common_names import DeviceName
from pitop.core.exceptions import UnavailableComponent
from pitop.system import device_type


class SupportsMiniscreen:
    def __init__(self):
        from pitop.miniscreen import Miniscreen

        self._miniscreen = None
        if device_type() == DeviceName.pi_top_4.value:
            self._miniscreen = Miniscreen()

    @property
    def miniscreen(self):
        if self._miniscreen:
            return self._miniscreen
        raise UnavailableComponent("Miniscreen isn't available on this device")

    @property
    def oled(self):
        return self.miniscreen



# coding=utf-8
# pynput
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""This module contains the base implementation.

The actual interface to mouse classes is defined here, but the
implementation is located in a platform dependent module.
"""

# pylint: disable=R0903
# We implement stubs

import enum

from pynput import _logger
from pynput._util import AbstractListener


class Button(enum.Enum):
    """The various buttons.

    The actual values for these items differ between platforms. Some
    platforms may have additional buttons, but these are guaranteed to
    be present everywhere.
    """

    #: An unknown button was pressed
    unknown = 0

    #: The left button
    left = 1

    #: The middle button
    middle = 2

    #: The right button
    right = 3


class Controller(object):
    """A controller for sending virtual mouse events to the system."""

    def __init__(self):
        self._log = _logger(self.__class__)

    @property
    def position(self):
        """The current position of the mouse pointer.

        This is the tuple ``(x, y)``, and setting it will move the
        pointer.
        """
        return self._position_get()

    @position.setter
    def position(self, pos):
        self._position_set(pos)

    def scroll(self, dx, dy):
        """Sends scroll events.

        :param int dx: The horizontal scroll. The units of scrolling is
            undefined.

        :param int dy: The vertical scroll. The units of scrolling is
            undefined.

        :raises ValueError: if the values are invalid, for example out of
            bounds
        """
        self._scroll(dx, dy)

    def press(self, button):
        """Emits a button press event at the current position.

        :param Button button: The button to press.
        """
        self._press(button)

    def release(self, button):
        """Emits a button release event at the current position.

        :param Button button: The button to release.
        """
        self._release(button)

    def move(self, dx, dy):
        """Moves the mouse pointer a number of pixels from its current
        position.

        :param int x: The horizontal offset.

        :param int dy: The vertical offset.

        :raises ValueError: if the values are invalid, for example out of
            bounds
        """
        self.position = tuple(sum(i) for i in zip(self.position, (dx, dy)))

    def click(self, button, count=1):
        """Emits a button click event at the current position.

        The default implementation sends a series of press and release events.

        :param Button button: The button to click.

        :param int count: The number of clicks to send.
        """
        with self as controller:
            for _ in range(count):
                controller.press(button)
                controller.release(button)

    def __enter__(self):
        """Begins a series of clicks.

        In the default :meth:`click` implementation, the return value of this
        method is used for the calls to :meth:`press` and :meth:`release`
        instead of ``self``.

        The default implementation is a no-op.
        """
        return self

    def __exit__(self, exc_type, value, traceback):
        """Ends a series of clicks."""
        pass

    def _position_get(self):
        """The implementation of the getter for :attr:`position`.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def _position_set(self, pos):
        """The implementation of the setter for :attr:`position`.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def _scroll(self, dx, dy):
        """The implementation of the :meth:`scroll` method.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def _press(self, button):
        """The implementation of the :meth:`press` method.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def _release(self, button):
        """The implementation of the :meth:`release` method.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()


# pylint: disable=W0223; This is also an abstract class
class Listener(AbstractListener):
    """A listener for mouse events.

    Instances of this class can be used as context managers. This is equivalent
    to the following code::

        listener.start()
        try:
            listener.wait()
            with_statements()
        finally:
            listener.stop()

    This class inherits from :class:`threading.Thread` and supports all its
    methods. It will set :attr:`daemon` to ``True`` when created.

    :param callable on_move: The callback to call when mouse move events occur.

        It will be called with the arguments ``(x, y)``, which is the new
        pointer position. If this callback raises :class:`StopException` or
        returns ``False``, the listener is stopped.

    :param callable on_click: The callback to call when a mouse button is
        clicked.

        It will be called with the arguments ``(x, y, button, pressed)``,
        where ``(x, y)`` is the new pointer position, ``button`` is one of the
        :class:`Button` values and ``pressed`` is whether the button was
        pressed.

        If this callback raises :class:`StopException` or returns ``False``,
        the listener is stopped.

    :param callable on_scroll: The callback to call when mouse scroll
        events occur.

        It will be called with the arguments ``(x, y, dx, dy)``, where
        ``(x, y)`` is the new pointer position, and ``(dx, dy)`` is the scroll
        vector.

        If this callback raises :class:`StopException` or returns ``False``,
        the listener is stopped.

    :param bool suppress: Whether to suppress events. Setting this to ``True``
        will prevent the input events from being passed to the rest of the
        system.

    :param kwargs: Any non-standard platform dependent options. These should be
        prefixed with the platform name thus: ``darwin_``, ``xorg_`` or
        ``win32_``.

        Supported values are:

        ``darwin_intercept``
            A callable taking the arguments ``(event_type, event)``, where
            ``event_type`` is any mouse related event type constant, and
            ``event`` is a ``CGEventRef``.

            This callable can freely modify the event using functions like
            ``Quartz.CGEventSetIntegerValueField``. If this callable does not
            return the event, the event is suppressed system wide.

        ``win32_event_filter``
            A callable taking the arguments ``(msg, data)``, where ``msg`` is
            the current message, and ``data`` associated data as a
            `MSLLHOOKSTRUCT <https://msdn.microsoft.com/en-us/library/windows/desktop/ms644970(v=vs.85).aspx>`_.

            If this callback returns ``False``, the event will not
            be propagated to the listener callback.

            If ``self.suppress_event()`` is called, the event is suppressed
            system wide.
    """

    def __init__(
        self, on_move=None, on_click=None, on_scroll=None, suppress=False, **kwargs
    ):
        self._log = _logger(self.__class__)
        prefix = self.__class__.__module__.rsplit(".", 1)[-1][1:] + "_"
        self._options = {
            key[len(prefix) :]: value
            for key, value in kwargs.items()
            if key.startswith(prefix)
        }
        super(Listener, self).__init__(
            on_move=on_move, on_click=on_click, on_scroll=on_scroll, suppress=suppress
        )


# pylint: enable=W0223



import atexit
import logging
from math import atan2, degrees, sqrt
from os import path
from threading import Event, Thread
from time import sleep, strftime, time
from warnings import catch_warnings, filterwarnings

import numpy as np

# Enables "add_subplot(projection='3d')"
from mpl_toolkits import mplot3d  # noqa: F401, lgtm[py/unused-import]
from scipy.linalg import sqrtm

from pitop.pma.common.math_functions.ellipsoid_functions import (
    get_ellipsoid_geometric_params,
    least_squares_ellipsoid_fit,
    plot_ellipsoid,
)
from pitop.pma.imu_controller import ImuController

logger = logging.getLogger(__name__)


class ImuCalibration:
    __MAG_POLL_FREQUENCY = 25.0  # Hz
    __GYRO_POLL_FREQUENCY = 5.0  # Hz
    __SLEEP_TIME = 0.005
    __MAG_FILTER_SIZE = 5

    def __init__(self):
        self.imu_controller = ImuController()
        self.imu_controller.acc_enable = True
        self.imu_controller.gyro_enable = True
        self.imu_controller.mag_enable = True

        self.__mag_measurements = np.zeros((1, 3), dtype=float)
        self.__mag_filter_array = np.zeros((self.__MAG_FILTER_SIZE, 3), dtype=float)
        self.__hard_iron_offset = None
        self.__soft_iron_matrix = None
        self.__mag_measurements_calibrated = None
        self.__center = None
        self.__radii = None
        self.__rotation_matrix = None
        self.__field_strength = None
        self.__test_data = None
        self.__save_data_name = None
        atexit.register(self.imu_controller.cleanup)

    @property
    def mag_data(self):
        return self.__mag_measurements

    def calibrate_magnetometer(
        self, test_data=None, save_data_name=None, update_mcu=True
    ):
        print("Starting magnetometer calibration")

        self.__test_data = test_data
        self.__save_data_name = save_data_name
        if test_data is None:
            self.__get_test_data()
        else:
            self.__mag_measurements = test_data

        if save_data_name and test_data is None:
            print("Saving test data...")
            with open(save_data_name, "wb") as f:
                np.save(f, self.__mag_measurements[self.__MAG_FILTER_SIZE :])

        print("Calculating calibration parameters...")

        self.__field_strength = self.__get_field_strength()

        M, n, d = self.__get_ellipse_parameters()

        (
            self.__center,
            self.__radii,
            self.__rotation_matrix,
        ) = get_ellipsoid_geometric_params(M, n, d)

        (
            self.__hard_iron_offset,
            self.__soft_iron_matrix,
        ) = self.__get_calibration_matrices(M, n, d, self.__field_strength)

        print("_hard_iron_offset: {}".format(self.__hard_iron_offset))
        print("_soft_iron_matrix: {}".format(self.__soft_iron_matrix))

        self.__mag_measurements_calibrated = self.__calibrate_mag_data()

        if update_mcu:
            self.imu_controller.write_mag_cal_params(
                self.__hard_iron_offset, self.__soft_iron_matrix
            )

    def plot_graphs(self, path_to_save_figure="/tmp/"):
        import matplotlib.pyplot as plt

        if path_to_save_figure and not path.isdir(path_to_save_figure):
            print(f"Provided path {path_to_save_figure} is not a valid directory")
            return

        x_uncal = self.mag_data[:, 0]
        y_uncal = self.mag_data[:, 1]
        z_uncal = self.mag_data[:, 2]

        x_cal = self.__mag_measurements_calibrated[:, 0]
        y_cal = self.__mag_measurements_calibrated[:, 1]
        z_cal = self.__mag_measurements_calibrated[:, 2]

        fig1 = plt.figure(1, figsize=(10, 10), dpi=80)
        fig1.suptitle(
            "Raw Magnetometer Data with Least Squares Ellipsoid Fit", fontsize=16
        )
        ax1 = fig1.add_subplot(111, projection="3d")
        ax1.axis("equal")
        ax1.set_xlabel("X")
        ax1.set_ylabel("Y")
        ax1.set_zlabel("Z")
        ax1.scatter(x_uncal, y_uncal, z_uncal, color="r")
        plot_ellipsoid(
            self.__center, self.__radii, self.__rotation_matrix, ax=ax1, plotAxes=True
        )

        fig2 = plt.figure(2, figsize=(10, 10), dpi=80)
        fig2.suptitle(
            "Calibrated Magnetometer Data with Field Strength Unit Sphere", fontsize=16
        )
        ax2 = fig2.add_subplot(111, projection="3d")
        ax2.axis("equal")
        ax2.set_xlabel("X")
        ax2.set_ylabel("Y")
        ax2.set_zlabel("Z")
        ax2.scatter(x_cal, y_cal, z_cal, color="r")

        # # plot unit sphere
        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x = np.outer(np.cos(u), np.sin(v)) * self.__field_strength
        y = np.outer(np.sin(u), np.sin(v)) * self.__field_strength
        z = np.outer(np.ones(np.size(u)), np.cos(v)) * self.__field_strength
        ax2.plot_wireframe(x, y, z, rstride=10, cstride=10, alpha=0.5)
        ax2.plot_surface(x, y, z, alpha=0.3, color="b")

        path_to_file = path.join(
            path_to_save_figure,
            "magnetometer_calibration_" + strftime("%Y-%m-%d-%H-%M-%S") + ".png",
        )
        plt.savefig(path_to_file)
        print(f"Calibration data was save to {path_to_file}")

    def __get_test_data(self):
        imu_controller = ImuController()
        thread_event = Event()
        mag_poll_thread = Thread(
            target=self.__poll_magnetometer_data,
            args=[
                thread_event,
                imu_controller,
            ],
            daemon=True,
        )
        mag_poll_thread.start()

        sleep(1)

        print("Hold the pi-top flat in the air so roll and pitch angles are zero.")
        self.__orientation_check(axis="z")
        sleep(1)
        print("Now rotate the pi-top 360 degrees whilst keeping it flat.")
        self.__rotation_check(axis="z")
        print("Done!")
        sleep(1)

        print(
            "Now turn the pi-top on it's side so the roll angle is +90 or -90 degrees."
        )
        self.__orientation_check(axis="x")
        sleep(1)
        print("Now rotate the pi-top 360 degrees whilst keeping it on its side.")
        self.__rotation_check(axis="x")
        print("Done!")
        sleep(1)

        print(
            "Now turn the pi-top on it's other side so the pitch angle is +90 or -90 degrees."
        )
        self.__orientation_check(axis="y")
        sleep(1)
        print("Now rotate the pi-top 360 degrees whilst keeping it on its side.")
        self.__rotation_check(axis="y")
        print("Done!")
        sleep(1)

        print("Now rotate/spin the pi-top in as many directions as possible.")
        sleep(1)
        input("Press any key when finished")
        print("Done!")
        sleep(1)

        thread_event.set()
        mag_poll_thread.join()

        sleep(1)

    def __rotation_check(self, axis: str):
        prev_time = time()
        degrees_rotated = 0
        while abs(degrees_rotated) < 360.0:
            current_time = time()
            dt = current_time - prev_time
            if dt > (1 / self.__GYRO_POLL_FREQUENCY):
                x, y, z = self.imu_controller.gyroscope_raw
                if axis == "x":
                    degrees_rotated += x * dt
                elif axis == "y":
                    degrees_rotated += y * dt
                elif axis == "z":
                    degrees_rotated += z * dt
                # print("degrees_rotated: {}".format(degrees_rotated))
                prev_time = current_time
            else:
                sleep(self.__SLEEP_TIME)

    def __orientation_check(self, axis: str):
        if axis not in ("x", "y", "z"):
            raise ValueError("axis value must me 'x', 'y' or 'z'.")

        def check_x_axis(roll_value, pitch_value):
            if abs(roll_value) > 85 and abs(pitch_value) < 5:
                print("---------- Correct orientation found ----------")
                return True

        def check_y_axis(roll_value, pitch_value):
            if abs(roll_value) < 5 and abs(pitch_value) > 85:
                print("---------- Correct orientation found ----------")
                return True

        def check_z_axis(roll_value, pitch_value):
            if abs(roll_value) < 5 and abs(pitch_value) < 5:
                print("---------- Correct orientation found ----------")
                return True

        if axis == "x":
            roll_check = check_x_axis
        elif axis == "y":
            roll_check = check_y_axis
        else:
            roll_check = check_z_axis

        while True:
            roll, pitch = self.accelerometer_orientation()
            if roll_check(roll, pitch):
                break
            else:
                sleep(self.__SLEEP_TIME)

    def accelerometer_orientation(self):
        x, y, z = self.imu_controller.accelerometer_raw
        roll = degrees(atan2(x, sqrt(y ** 2 + z ** 2)))
        pitch = degrees(atan2(-y, sqrt(x ** 2 + z ** 2)))

        return roll, pitch

    def __poll_magnetometer_data(self, thread_event, imu_controller):
        print("Polling mag data...")
        sleep(1)
        prev_time = time()
        error_tolerance = 0.01
        error_count = 0
        while not thread_event.is_set():
            current_time = time()
            if current_time - prev_time > (1 / self.__MAG_POLL_FREQUENCY):
                x, y, z = imu_controller.magnetometer_raw
                if (
                    abs(x) < error_tolerance
                    and abs(y) < error_tolerance
                    and abs(z) < error_tolerance
                ):
                    print("Read error, trying again...")
                    error_count += 1
                    if error_count > 50:
                        logger.error(
                            "Cannot get magnetometer readings, please try re-docking your pi-top into the "
                            "Expansion Plate and running the calibrator again."
                        )
                        exit()
                    continue
                new_mag_data = [x, y, z]
                self.__mag_filter_array, mag_median = self.__running_median(
                    self.__mag_filter_array, new_mag_data
                )
                self.__mag_measurements = np.append(
                    self.__mag_measurements, [mag_median], axis=0
                )
                prev_time = current_time
            else:
                sleep(self.__SLEEP_TIME)

    @staticmethod
    def __running_median(old_array, new_data):
        new_array = np.append(np.delete(old_array, 0, 0), [new_data], axis=0)
        new_median = np.median(new_array, axis=0)
        return new_array, new_median

    def __get_field_strength(self):
        # from  https://www.nxp.com/docs/en/application-note/AN4246.pdf
        squared_measurements = np.square(self.__mag_measurements)

        matrix_y = np.sum(squared_measurements, axis=1)

        rows, columns = self.__mag_measurements.shape
        ones = np.ones((rows, 1))
        matrix_x = np.column_stack((self.__mag_measurements, ones))

        matrix_x_transpose = np.transpose(matrix_x)

        x_t_x_inverse = np.linalg.inv(np.matmul(matrix_x_transpose, matrix_x))
        x_t_y = np.matmul(matrix_x_transpose, matrix_y)

        beta_vector = np.matmul(x_t_x_inverse, x_t_y)

        correction_vector = 0.5 * beta_vector[0:3]

        field_strength = sqrt(beta_vector[3] + np.sum(np.square(correction_vector)))

        return field_strength

    def __get_ellipse_parameters(self):
        x_uncal = self.mag_data[:, 0]
        y_uncal = self.mag_data[:, 1]
        z_uncal = self.mag_data[:, 2]

        M, n, d = least_squares_ellipsoid_fit(x_uncal, y_uncal, z_uncal)

        return M, n, d

    def __get_calibration_matrices(self, M, n, d, field_strength):
        hard_iron_offset = None
        soft_iron_matrix = None

        with catch_warnings():
            filterwarnings("error")
            try:
                Minv = np.linalg.inv(M)
                soft_iron_matrix = np.real(
                    field_strength
                    / np.sqrt(np.dot(n.T, np.dot(Minv, n)) - d)
                    * sqrtm(M)
                )
                hard_iron_offset = -np.dot(Minv, n)
            except Warning as e:
                logger.error("Calibration error: {}".format(e))
                if self.__test_data is None:
                    logger.info("Starting calibration process again...")
                    sleep(3)
                    self.calibrate_magnetometer(
                        test_data=self.__test_data, save_data_name=self.__save_data_name
                    )
                else:
                    logger.info("Please try again with different test data.")
                    exit()

        return hard_iron_offset, soft_iron_matrix

    def __calibrate_mag_data(self):
        x_uncal = self.mag_data[:, 0]
        y_uncal = self.mag_data[:, 1]
        z_uncal = self.mag_data[:, 2]
        x_cal = np.zeros(x_uncal.shape)
        y_cal = np.zeros(x_uncal.shape)
        z_cal = np.zeros(x_uncal.shape)

        error_sum = 0
        error_squared_sum = 0
        for i in range(len(x_uncal)):
            h_actual = np.array([[x_uncal[i], y_uncal[i], z_uncal[i]]]).T
            h_estimate = np.matmul(
                self.__soft_iron_matrix, h_actual - self.__hard_iron_offset
            )
            x_cal[i] = h_estimate[0]
            y_cal[i] = h_estimate[1]
            z_cal[i] = h_estimate[2]
            magnitude = np.sqrt(np.dot(h_estimate.T, h_estimate))
            error = magnitude[0][0] - self.__field_strength
            error_squared = error ** 2
            error_sum += error
            error_squared_sum += error_squared
        print("Average Error: {}".format(error_sum / x_uncal.shape))
        print("Error Variance: {}".format(error_squared_sum / x_uncal.shape))

        mag_measurements_cal = np.hstack(
            (np.array([x_cal]).T, np.array([y_cal]).T, np.array([z_cal]).T)
        )

        return mag_measurements_cal



import atexit
from os import getenv
from typing import Union

from imutils.video import FPS

from pitop.core import ImageFunctions
from pitop.processing.algorithms.faces.core.face import Face
from pitop.processing.core.load_models import load_face_landmark_predictor
from pitop.processing.core.vision_functions import (
    import_dlib,
    import_face_utils,
    import_imutils,
    import_opencv,
    tuple_for_color_by_name,
)

cv2 = None
dlib = None
imutils = None
face_utils = None


class FaceDetector:
    _FACE_DETECTOR_PYRAMID_LAYERS = (
        1  # set higher to detect smaller faces. Cost: large reduction in detection FPS.
    )

    def __init__(
        self,
        image_processing_width: Union[int, None] = 320,
        format: str = "OpenCV",
        enable_tracking: bool = True,
        dlib_landmark_predictor_filename: str = "shape_predictor_68_face_landmarks.dat",
    ):
        """
        :param Union[int, None] image_processing_width: image width to scale to for image processing, set to None for
        no scaling.
        :param str format: desired output image format.
        :param bool enable_tracking: enable dlib's correlaction tracker to track the detected face between frames.
        :param str dlib_landmark_predictor_filename: Filename for facial features predictor. Use 5 landmark version for
        slightly better performance whilst retaining ability to calculate face angle. May be incompatible with further
         processing e.g. emotion detection requires the 68-landmark version.
        """
        self.__import_libs()

        self._image_processing_width = image_processing_width
        self._format = format
        self._face_rectangle_detector = dlib.get_frontal_face_detector()
        self._predictor = load_face_landmark_predictor(
            model_filename=dlib_landmark_predictor_filename
        )
        self._clahe_filter = cv2.createCLAHE(clipLimit=5)
        self._frame_scaler = None
        self.face = Face()
        self._enable_tracking = enable_tracking
        self._face_tracker = None

        # Enable FPS if environment variable is set
        self._print_fps = getenv("PT_ENABLE_FPS", "0") == "1"
        if self._print_fps:
            self._fps = FPS().start()
            atexit.register(self.__print_fps)

    def __call__(self, frame):
        """Detect a face in an image frame.

        :param frame: Image frame in OpenCV or PIL format.
        :return: Face object containing data about the detected face.
        """
        self.__import_libs()
        frame = ImageFunctions.convert(frame, format="OpenCV")

        if self._frame_scaler is None:
            _, width = frame.shape[0:2]
            self._frame_scaler = (
                width / self._image_processing_width
                if self._image_processing_width is not None
                else 1
            )

        frame_to_process = self.__get_frame_to_process(frame=frame)

        if self._face_tracker is None:
            # if the face tracker has not been started, we first need to use face detection to find a face
            face_rectangle, face_center, face_features = self.__detect_largest_face(
                frame=frame_to_process
            )
            if face_center is not None and self._enable_tracking:
                # Face found, enable dlib correlation tracker for subsequent calls
                self.__start_tracker(frame=frame_to_process, rectangle=face_rectangle)
        else:
            # We are in face tracking mode, use dlib correlation tracker to track the face
            face_rectangle, face_center, face_features = self.__track_face(
                frame=frame_to_process
            )
            if face_center is None:
                self.__stop_tracker()
                # attempt to detect face since tracker has failed
                face_rectangle, face_center, face_features = self.__detect_largest_face(
                    frame=frame_to_process
                )

        self.face = self.__prepare_face_data(
            frame=frame,
            face=self.face,
            rectangle=face_rectangle,
            center=face_center,
            features=face_features,
        )
        if self._print_fps:
            self._fps.update()

        return self.face

    def __import_libs(self):
        global cv2, dlib, imutils, face_utils
        if cv2 is None:
            cv2 = import_opencv()
        if dlib is None:
            dlib = import_dlib()
        if face_utils is None:
            face_utils = import_face_utils()
        if imutils is None:
            imutils = import_imutils()

    def __get_frame_to_process(self, frame):
        """Resize frame, convert to grayscale and use contrast limited adaptive
        histogram equalization (CLAHE) to improve the contrast of the image so
        that face features are more pronounced.

        :param frame: original frame from the camera in OpenCV format
        :return: OpenCV image to send to face processing algorithms
        """
        return self._clahe_filter.apply(
            cv2.cvtColor(
                imutils.resize(frame, width=self._image_processing_width),
                cv2.COLOR_BGR2GRAY,
            )
        )

    def __detect_largest_face(self, frame):
        """Use dlib face rectangle detector on the frame and process the
        detected rectangles to get the face data.

        :param frame: Image frame to process for faces
        :return: Detected face data
        """
        face_rectangle, face_center, face_features = self.__process_detected_rectangles(
            frame,
            self._face_rectangle_detector(frame, self._FACE_DETECTOR_PYRAMID_LAYERS),
        )

        return face_rectangle, face_center, face_features

    def __process_detected_rectangles(self, frame, rectangles_dlib):
        """Find largest face rectangle and process it to get the required face
        data.

        :param frame: OpenCV frame to use for processing
        :param rectangles_dlib: Rectangles found using dlib's face detector in the dlib Rectangle format.
        :return: Detected face data
        """
        if len(rectangles_dlib) == 0:
            return None, None, None

        largest_rectangle_dlib = None
        area = 0
        face_rectangle = None
        face_center = None
        for rectangle_dlib in rectangles_dlib:
            x, y, w, h = face_utils.rect_to_bb(rectangle_dlib)
            current_area = w * h
            if current_area > area:
                area = current_area
                largest_rectangle_dlib = rectangle_dlib
                face_rectangle = (x, y, w, h)
                face_center = (int(x + w / 2), int(y + h / 2))

        face_features = self.__get_dlib_face_features(frame, largest_rectangle_dlib)

        return face_rectangle, face_center, face_features

    def __get_dlib_face_features(self, frame, dlib_rectangle):
        """Find's dlib face features using the predictor and then convert them
        to numpy format.

        :param frame: frame used to detect/track face rectangle.
        :param dlib_rectangle: Face rectangle in dlib Rectangle format.
        :return: 68x2 numpy array of x, y coordinates for facial features.
        """
        face_features_dlib = self._predictor(frame, dlib_rectangle)
        return face_utils.shape_to_np(face_features_dlib)

    def __start_tracker(self, frame, rectangle):
        self._face_tracker = dlib.correlation_tracker()
        self._face_tracker.start_track(
            frame,
            dlib.rectangle(
                rectangle[0],
                rectangle[1],
                rectangle[0] + rectangle[2],
                rectangle[1] + rectangle[3],
            ),
        )

    def __stop_tracker(self):
        self._face_tracker = None

    def __track_face(self, frame):
        """Use dlib correlation tracker to track the face from the last frame
        in the current frame.

        :param frame: OpenCV frame to use for processing
        :return: Face data
        """

        def convert_dlib_rect_to_int_type(rectangle):
            x_start = int(round(rectangle.left()))
            y_start = int(round(rectangle.top()))
            x_end = int(round(rectangle.right()))
            y_end = int(round(rectangle.bottom()))
            return dlib.rectangle(x_start, y_start, x_end, y_end)

        peak_to_side_lobe_ratio = self._face_tracker.update(frame)

        if peak_to_side_lobe_ratio < 7.0:
            # Object occluded or lost when PSR found to be below a certain threshold (7.0 according to Bolme et al)
            # Treat as failed attempt to track face and return
            return None, None, None

        rectangle_dlib = convert_dlib_rect_to_int_type(
            self._face_tracker.get_position()
        )
        face_features = self.__get_dlib_face_features(frame, rectangle_dlib)

        face_rectangle = face_utils.rect_to_bb(rectangle_dlib)
        face_center = (
            face_rectangle[0] + int(round(face_rectangle[2] / 2)),
            face_rectangle[1] + int(round(face_rectangle[3] / 2)),
        )

        return face_rectangle, face_center, face_features

    def __prepare_face_data(self, frame, face, rectangle, center, features):
        face.original_detection_frame = frame

        if center is None:
            face.clear()
            face.robot_view = ImageFunctions.convert(frame, format=self._format)
            return face

        # resize back to original frame resolution
        face.rectangle = tuple((int(item * self._frame_scaler) for item in rectangle))
        face.center_default = tuple((int(item * self._frame_scaler) for item in center))
        face.features = (features * self._frame_scaler).astype("int")

        face.robot_view = ImageFunctions.convert(
            self.__draw_on_frame(frame=frame.copy(), face=face), format=self._format
        )
        return face

    @staticmethod
    def __draw_on_frame(frame, face):
        x, y, w, h = face.rectangle
        cv2.rectangle(
            frame,
            (x, y),
            (x + w, y + h),
            tuple_for_color_by_name("dodgerblue", bgr=True),
            2,
        )

        cv2.drawMarker(
            frame,
            face.center_default,
            tuple_for_color_by_name("orangered", bgr=True),
            markerType=cv2.MARKER_CROSS,
            markerSize=10,
            thickness=3,
            line_type=cv2.FILLED,
        )

        return frame

    def __print_fps(self):
        self._fps.stop()
        print(f"[INFO] Elapsed time: {self._fps.elapsed():.2f}")
        print(f"[INFO] Approx. FPS: {self._fps.fps():.2f}")



from pitop.processing.core.vision_functions import center_reposition

from .face_utils import (
    get_face_angle,
    left_eye_center,
    left_eye_dimensions,
    mouth_center,
    mouth_dimensions,
    nose_bottom,
    pupil_distance,
    right_eye_center,
    right_eye_dimensions,
)


class Face:
    def __init__(self):
        self._center = None
        self._features = None
        self._rectangle = None
        self._robot_view = None
        self._original_frame = None

    def clear(self):
        self.center_default = None
        self.features = None
        self.rectangle = None

    @property
    def center(self):
        """
        :return: Face center with the coordinate system in the middle of the frame.
                 Positive x points right, positive y points up.
        :type: tuple
        """
        return (
            center_reposition(self.center_default, self.original_detection_frame)
            if self.found
            else None
        )

    @property
    def center_default(self):
        """
        :return: Face center with the coordinate system in the top left of the frame.
                 Positive x points right, positive y points down. This is the same coordinate system used by OpenCV
                 and PIL.
        :type: tuple
        """
        return self._center

    @center_default.setter
    def center_default(self, value):
        self._center = value

    @property
    def angle(self):
        return get_face_angle(self.features) if self.found else None

    @property
    def pupil_distance(self):
        return pupil_distance(self.features) if self.found else None

    @property
    def left_eye_center(self):
        return left_eye_center(self.features) if self.found else None

    @property
    def left_eye_dimensions(self):
        return left_eye_dimensions(self.features) if self.found else None

    @property
    def right_eye_center(self):
        return right_eye_center(self.features) if self.found else None

    @property
    def right_eye_dimensions(self):
        return right_eye_dimensions(self.features) if self.found else None

    @property
    def mouth_center(self):
        return mouth_center(self.features) if self.found else None

    @property
    def mouth_dimensions(self):
        return mouth_dimensions(self.features) if self.found else None

    @property
    def nose_bottom(self):
        return nose_bottom(self.features) if self.found else None

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, value):
        self._features = value

    @property
    def rectangle(self):
        return self._rectangle

    @rectangle.setter
    def rectangle(self, value):
        self._rectangle = value

    @property
    def robot_view(self):
        return self._robot_view

    @robot_view.setter
    def robot_view(self, value):
        self._robot_view = value

    @property
    def original_detection_frame(self):
        return self._original_frame

    @original_detection_frame.setter
    def original_detection_frame(self, value):
        self._original_frame = value

    @property
    def found(self) -> bool:
        """
        :return: Boolean to determine if a face was found in the frame.
        :rtype: bool
        """
        return self.center_default is not None



from os import listdir

from PIL import Image

from pitop.core.ImageFunctions import convert
from pitop.core.import_opencv import import_opencv

valid_rotate_angles = [-270, -180, -90, 0, 90, 180, 270]


class UsbCamera:
    def __init__(
        self,
        index: int = None,
        resolution=None,
        rotate_angle: int = 0,
        flip_top_bottom: bool = False,
        flip_left_right: bool = False,
    ):

        VideoCapture = import_opencv().VideoCapture

        self.__camera = None
        self.index = -1 if index is None else index

        self._flip_top_bottom = flip_top_bottom
        self._flip_left_right = flip_left_right

        if rotate_angle not in valid_rotate_angles:
            raise ValueError(
                f"Rotate angle must be one of "
                f"{', '.join([str(x) for x in valid_rotate_angles[:-1]])} or "
                f"{str(valid_rotate_angles[-1])}"
            )
        else:
            self._rotate_angle = rotate_angle

        def create_camera_object(index, resolution=None):
            cap = VideoCapture(index)
            if resolution is not None:
                cap.set(3, resolution[0])
                cap.set(4, resolution[1])
            return cap

        self.__camera = create_camera_object(self.index, resolution)
        if not self.is_opened():
            self.__camera = None
            raise IOError(
                "Error opening camera. Make sure it's correctly connected via USB."
            ) from None

    def __del__(self):
        try:
            if hasattr(self.__camera, "release"):
                self.__camera.release()
        except AttributeError:
            # Camera was not initialized
            pass

    def get_frame(self):
        if not self.is_opened():
            raise IOError("Camera not connected")

        result, frame = self.__camera.read()
        if not result:
            raise ValueError("Couldn't grab frame from camera")

        # Always PIL format
        pil_image = convert(frame, "PIL").rotate(angle=self._rotate_angle, expand=True)

        if self._flip_top_bottom:
            pil_image = pil_image.transpose(method=Image.FLIP_TOP_BOTTOM)

        if self._flip_left_right:
            pil_image = pil_image.transpose(method=Image.FLIP_LEFT_RIGHT)

        return pil_image

    def is_opened(self):
        return self.__camera is not None and self.__camera.isOpened()

    @staticmethod
    def list_device_indexes():
        VideoCapture = import_opencv().VideoCapture

        # find "video" device indexes in /dev
        device_names = [name for name in listdir("/dev") if "video" in name]
        device_indexes = [int(dev[len("video") :]) for dev in device_names]
        # indexes >= 10 are for bcm2835, not useful for us
        camera_indexes = [i for i in device_indexes if i < 10]
        camera_indexes.sort()

        # Not all devices actually provide video
        # eg: video1 can be a metadata device for video0
        indexes = []
        for index in camera_indexes:
            try:
                cap = VideoCapture(index)
                if cap.read()[0]:
                    indexes.append(index)
                    cap.release()
            except Exception:
                continue

        return indexes



from argparse import ArgumentParser
from sys import exit, stderr

# This file serves as support for deprecated CLI paths, that will be removed in the future.


class dotdict(dict):
    """dot.notation access to dictionary attributes."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def __run(cli_cls, args):
    try:
        cli_object = cli_cls(args)
        exit_code = cli_object.run()
    except Exception as e:
        print(f"Error: {e}")
        exit_code = 1
    exit(exit_code)


def run(cli_cls):
    """Runs the CLI provided as argument, using the class properties and
    methods.

    Args:
        cli_cls (CliBaseClass): CLI class constructor. Must inherit from CliBaseClass.
    """
    cli_name = cli_cls.cli_name
    print(
        f"Note: Use of 'pt-{cli_name}' is now deprecated. "
        f"Please use 'pi-top {cli_name}' instead.",
        file=stderr,
    )

    parser = ArgumentParser(prog=f"pt-{cli_name}", description=cli_cls.parser_help)
    cli_cls.add_parser_arguments(parser)
    args = parser.parse_args()
    __run(cli_cls, args)


def run_with_args(cli_cls, old_command, new_command, args_dict):
    """Runs a CLI command using the provided 'cli_cls' and the arguments in
    'args_dict'. This method doesn't use ArgumentParser, since it directly
    executes the CLI command using static arguments provided in 'args_dict'.

    Args:
        cli_cls (CliBaseClass): CLI class constructor. Must inherit from CliBaseClass.
        old_command (str): string with the deprecated CLI command
        new_command (str): string with the new CLI command
        args_dict (dict): dictionary with the arguments and values assigned to run the CLI
    """
    print(
        f"Note: Use of '{old_command}' is now deprecated. "
        f"Please use '{new_command}' instead.",
        file=stderr,
    )
    __run(cli_cls, dotdict(args_dict))



import math
from collections import deque
from enum import IntEnum

import numpy as np

from pitop.core.mixins import Stateful
from pitop.robotics.navigation.core.utils import normalize_angle

from ...filterpy.kalman import KalmanFilter


class State(IntEnum):
    x = 0
    y = 1
    theta = 2
    v = 3
    w = 4


class VelocityType(IntEnum):
    """Helper class to ensure v and w values are accessed at the correct
    position in various vectors and matrices."""

    v = 0
    w = 1


class VelocityMeasurements(IntEnum):
    previous = 0
    current = 1


class StateFilter(Stateful):
    _sigma_default_dt = 0.1

    def __init__(self, measurement_frequency, wheel_separation):
        self._kalman_filter = KalmanFilter(dim_x=len(State), dim_z=2, dim_u=2)
        self._velocities = deque(maxlen=2)
        self._velocities.append(
            np.zeros((len(VelocityType), 1), dtype=float)
        )  # [v, w].

        # Starting covariance is small since we know with high confidence we are starting at [0, 0, 0] with 0 velocity
        self._kalman_filter.P = np.eye(5) * 1e-6

        # Motor encoder has 115 RPM resolution. This is 2.75 RPM after dividing by gear ratio.
        # Converting to m/s this is 0.01 m/s resolution (from wheel diameter of 0.0718)
        # Divide by two to give a max error of +/- 0.005 m/s
        # This would usually be taken as the 3*sigma value but given all the other errors in the system (wheel diameter,
        # slippage etc) we'll use this value directly for sigma
        linear_velocity_sigma = 0.005

        # the Q matrix is the covariance of the expected state change over the time interval dt
        Q_sigma = linear_velocity_sigma / (
            self._sigma_default_dt * measurement_frequency
        )
        self._kalman_filter.Q = np.diag(
            [
                Q_sigma ** 2,
                Q_sigma ** 2,
                math.radians(Q_sigma),
                Q_sigma ** 2,
                Q_sigma ** 2,
            ]
        )

        # State transition function for x1 = Fx0 + Bu0
        self._kalman_filter.F = np.diag([1, 1, 1, 0, 0])

        self._odom_linear_velocity_variance = linear_velocity_sigma ** 2
        # angular velocity is calculated from motor velocities, maximum error is 0.005 * 2 across both wheel speeds
        # divide by wheel separation to get resulting standard deviation for angular velocity
        self._odom_angular_velocity_variance = (
            linear_velocity_sigma * 2 / wheel_separation
        ) ** 2

        # Gyroscope resolution from datasheet is 1/131 = 0.0076 degrees/second +/- 1.5%
        # Measure variance from stationary IMU is sigma**2 = 0.0018 over approx 200 samples
        # IMU is likely to have a bias term associated and is less trustworthy than the odometry, use a larger value
        # for now.
        self._imu_angular_velocity_variance = np.radians(0.5 ** 2)

        Stateful.__init__(self, children=[])

    def __str__(self):
        degree_symbol = "\N{DEGREE SIGN}"
        return (
            f"               x = {self.x:.3} m (+/- {self.x_tolerance:.3})\n"
            f"               y = {self.y:.3} m (+/- {self.y_tolerance:.3})\n"
            f"           Angle = {self.angle:.3}{degree_symbol} (+/- {self.angle_tolerance:.3})\n"
            f"        Velocity = {self.v:.3} m/s (+/- {self.v_tolerance:.3})\n"
            f"Angular velocity = {math.degrees(self.w):.3} {degree_symbol}/s "
            f"(+/- {math.degrees(self.w_tolerance):.3})\n"
        )

    @property
    def own_state(self):
        return {
            "x_position": self.x,
            "y_position": self.y,
            "angle_position": self.angle,
            "linear_velocity": self.v,
            "angular_velocity": self.w,
        }

    def add_measurements(self, odom_measurements, dt, imu_measurements=None):
        self._velocities.append(odom_measurements)
        self.__kalman_predict(u=self._velocities[VelocityMeasurements.previous], dt=dt)
        self.__kalman_update(z_imu=imu_measurements)
        # normalize state angle to lie between -pi and +pi
        self.angle_rad = normalize_angle(self.angle_rad)

    def __kalman_predict(self, u, dt):
        """
        Full predict equation is x1 = Fx0 + Bu0 where F is constant and defined in __init__
        B is the control transition matrix which is multiplied by u0 (control vector)
        B is derived from basic newtonian mechanics to predict new state from the control input - in our case, we are
        using the previously measured velocities as the control input

        Predict equations:
        x1 = x0 + dt * v0 * cos(theta + 0.5 * dt * w0)
        y1 = y0 + dt * v0 * sin(theta + 0.5 * dt * w0)
            (the 0.5 is a first order approximation to a curved path resulting from angular velocity w0)
        theta1 = dt * w0
        v1 = v0  # zero acceleration model, process noise Q allows for perturbations in velocity
        w1 = w0  # zero acceleration model, process noise Q allows for perturbations in velocity

        :param u: Control input vector - we take the previous velocity measurements for this since they are a better
        predictor than the actual control command sent to the Expansion Plate MCU (this is very commonly done).
        :param dt: time difference between measurements.
        """
        B = np.array(
            [
                [dt * math.cos(self.angle_rad + 0.5 * dt * self.w), 0],
                [dt * math.sin(self.angle_rad + 0.5 * dt * self.w), 0],
                [0.0, dt],
                [1.0, 0.0],
                [0.0, 1.0],
            ]
        )
        self._kalman_filter.predict(u=u, B=B)

    def __kalman_update(self, z_imu):
        if z_imu is None:
            self._kalman_filter.update(z=None)
            return

        self._kalman_filter.dim_z = 1
        self._kalman_filter.H = np.array([[0, 0, 0, 0, 1]])
        self._kalman_filter.R = np.diag([self._imu_angular_velocity_variance])
        self._kalman_filter.update(z=z_imu)

    @property
    def x(self):
        """
        :return float: Estimated x position of the robot in meters.
        """
        return self._kalman_filter.x[State.x, 0]

    @x.setter
    def x(self, value):
        self._kalman_filter.x[State.x, 0] = value

    @property
    def x_tolerance(self):
        """Returns the 2-sigma value for the variance in x.

        About 95% of the possible values for the true value of x lie
        within two standard deviations of the mean.
        :return: 2-sigma value of the x position
        """
        return 2 * np.sqrt(self._kalman_filter.P[State.x, State.x])

    @property
    def y(self):
        """
        :return float: Estimated y position of the robot in meters.
        """
        return self._kalman_filter.x[State.y, 0]

    @y.setter
    def y(self, value):
        self._kalman_filter.x[State.y, 0] = value

    @property
    def y_tolerance(self):
        """Returns the 2-sigma value for the variance in y.

        About 95% of the possible values for the true value of y lie
        within two standard deviations of the mean.
        :return: 2-sigma value of the y position
        """
        return 2 * np.sqrt(self._kalman_filter.P[State.y, State.y])

    @property
    def angle(self):
        """
        :return float: Estimated angle of the robot in degrees.
        """
        return math.degrees(self.angle_rad)

    @angle.setter
    def angle(self, value):
        self.angle_rad = math.radians(value)

    @property
    def angle_tolerance(self):
        """Returns the 2-sigma value for the variance in the angle.

        About 95% of the possible values for the true value of
        the angle lie within two standard deviations of the mean.
        :return: 2-sigma value of the angle position in degrees
        """
        return math.degrees(self.angle_rad_tolerance)

    @property
    def angle_rad(self):
        return self._kalman_filter.x[State.theta, 0]

    @angle_rad.setter
    def angle_rad(self, value):
        self._kalman_filter.x[State.theta, 0] = value

    @property
    def angle_rad_tolerance(self):
        """Returns the 2-sigma value for the variance in the angle.

        About 95% of the possible values for the true value of
        the angle lie within two standard deviations of the mean.
        :return: 2-sigma value of the angle position in radians
        """
        return 2 * np.sqrt(self._kalman_filter.P[State.theta, State.theta])

    @property
    def v(self):
        """
        :return float: Estimated linear velocity of the robot in meters/second.
        """
        return self._kalman_filter.x[State.v, 0]

    @property
    def v_tolerance(self):
        return 2 * np.sqrt(self._kalman_filter.P[State.v, State.v])

    @property
    def w(self):
        """
        :return float: Estimated angular velocity of the robot in radians/second.
        """
        return self._kalman_filter.x[State.w, 0]

    @property
    def w_tolerance(self):
        return 2 * np.sqrt(self._kalman_filter.P[State.w, State.w])

    @property
    def position(self):
        """
        :return: Estimated position of the robot as a tuple in the form (x, y) where x and y are in meters.
        """
        return self.x, self.y

    @position.setter
    def position(self, value: tuple):
        self.x, self.y = value

    def reset_pose(self):
        self.x, self.y, self.angle_rad = (0, 0, 0)



from flask import Blueprint

from pitop.labs.web.blueprints.base import BaseBlueprint
from pitop.labs.web.blueprints.messaging import MessagingBlueprint
from pitop.labs.web.blueprints.video import VideoBlueprint
from pitop.labs.web.blueprints.webcomponents import WebComponentsBlueprint


class ControllerBlueprint(Blueprint):
    def __init__(self, get_frame=None, message_handlers={}, **kwargs):
        Blueprint.__init__(
            self,
            "controller",
            __name__,
            template_folder="templates",
            static_folder="static",
            **kwargs
        )

        self.base_blueprint = BaseBlueprint()
        self.components_blueprint = WebComponentsBlueprint()
        self.video_blueprint = VideoBlueprint(get_frame=get_frame)
        self.messaging_blueprint = MessagingBlueprint(message_handlers=message_handlers)

    def register(self, app, options, *args, **kwargs):
        # register child blueprints
        app.register_blueprint(self.base_blueprint, **options)
        app.register_blueprint(self.components_blueprint, **options)
        app.register_blueprint(self.video_blueprint, **options)
        app.register_blueprint(self.messaging_blueprint, **options)

        # register self
        Blueprint.register(self, app, options, *args, **kwargs)

    def broadcast(self, message):
        self.messaging_blueprint.broadcast(message)



from gpiozero import Button as gpiozero_Button

from pitop.core.mixins import Recreatable, Stateful
from pitop.pma.common import get_pin_for_port


class Button(Stateful, Recreatable, gpiozero_Button):
    """Encapsulates the behaviour of a push-button.

    A push-button is a simple switch mechanism for controlling some aspect of a circuit.

    :param str port_name: The ID for the port to which this component is connected
    """

    def __init__(self, port_name, name="button"):
        self._pma_port = port_name
        self.name = name

        Stateful.__init__(self)
        Recreatable.__init__(self, {"port_name": port_name, "name": self.name})
        gpiozero_Button.__init__(self, get_pin_for_port(self._pma_port))

    @property
    def own_state(self):
        return {
            "is_held": self.is_held,
            "is_pressed": self.is_pressed,
            "value": self.value,
        }

    def close(self):
        """Shut down the device and release all associated resources. This
        method can be called on an already closed device without raising an
        exception.

        This method is primarily intended for interactive use at the command
        line. It disables the device and releases its pin(s) for use by another
        device.

        You can attempt to do this simply by deleting an object, but unless
        you've cleaned up all references to the object this may not work (even
        if you've cleaned up all references, there's still no guarantee the
        garbage collector will actually delete the object at that point).  By
        contrast, the close method provides a means of ensuring that the object
        is shut down.

        For example, if you have a buzzer connected to port D0, but then wish
        to attach an LED instead:

            >>> from pitop import Buzzer, LED
            >>> bz = Buzzer("D0")
            >>> bz.on()
            >>> bz.off()
            >>> bz.close()
            >>> led = LED("D0")
            >>> led.blink()

        :class:`Device` descendents can also be used as context managers using
        the :keyword:`with` statement. For example:

            >>> from pitop import Buzzer, LED
            >>> with Buzzer("D0") as bz:
            ...     bz.on()
            ...
            >>> with LED("D0") as led:
            ...     led.on()
            ...
        """
        super(Button, self).close()



from distutils.util import strtobool
from os import uname
from time import strftime

from netifaces import AF_INET, AF_INET6, AF_LINK, ifaddresses, interfaces

from pitop.common.command_runner import run_command
from pitop.common.common_names import DeviceName
from pitop.common.pt_os import get_pitopOS_info
from pitop.common.sys_info import (
    get_debian_version,
    get_uname_release,
    get_uname_version,
)
from pitop.system import device_type, pitop_peripherals

from ..formatter import StdoutFormat, StdoutTable
from .hub_communication import HubCommunication
from .ptsoftware import PitopSoftware


def str_to_bool(value):
    return strtobool(value) == 0


class HealthCheck:
    RASPI_CONFIG_BOOT_SETTINGS = {
        "get_boot_cli": {
            "description": "Boot to Desktop?",
            "conversion_func": lambda x: not str_to_bool(x),
        },
        "get_autologin": {
            "description": "Autologin?",
            "conversion_func": str_to_bool,
        },
        "get_boot_wait": {
            "description": "Boot waits until a network connection is established?",
            "conversion_func": str_to_bool,
        },
        "get_boot_splash": {
            "description": "Show splash screen at boot?",
            "conversion_func": str_to_bool,
        },
    }

    RASPI_CONFIG_MISC = {
        "get_can_expand": {
            "description": "Filesystem can be expanded?",
            "conversion_func": str_to_bool,
        },
        "get_overscan": {
            "description": "Compensation for displays with overscan?",
            "conversion_func": str_to_bool,
        },
    }

    RASPI_CONFIG_INTERFACES = {
        "get_camera": {
            "description": "Camera interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_ssh": {
            "description": "SSH interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_vnc": {
            "description": "VNC interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_spi": {
            "description": "SPI interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_i2c": {
            "description": "I2C interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_serial": {
            "description": "Login shell accessible over serial?",
            "conversion_func": str_to_bool,
        },
        "get_serial_hw": {
            "description": "Serial interface is enabled?",
            "conversion_func": str_to_bool,
        },
        "get_onewire": {
            "description": "One-wire interface enabled?",
            "conversion_func": str_to_bool,
        },
        "get_rgpio": {
            "description": "GPIO server to be accessible over the network?",
            "conversion_func": str_to_bool,
        },
    }

    NETWORK_ENUM_LOOKUP = {AF_LINK: "LINK LAYER", AF_INET: "IPv4", AF_INET6: "IPv6"}

    VCGENMOD_SETTINGS = {
        "Throttled state of the system": "vcgencmd get_throttled",
        "SoC temperature": "vcgencmd measure_temp",
        "Clock frequency of ARM cores": "vcgencmd measure_clock arm",
        "Clock frequency of VC4 scaler cores": "vcgencmd measure_clock core",
        "Clock frequency of H264 block": "vcgencmd measure_clock H264",
        "Clock frequency of Image Signal Processor": "vcgencmd measure_clock isp",
        "Clock frequency of 3D block": "vcgencmd measure_clock v3d",
        "Clock frequency of UART": "vcgencmd measure_clock uart",
        "Clock frequency of PWM block (analog audio output)": "vcgencmd measure_clock pwm",
        "Clock frequency of SD card interface": "vcgencmd measure_clock emmc",
        "Clock frequency of Pixel valve": "vcgencmd measure_clock pixel",
        "Clock frequency of Analog Video Encoder": "vcgencmd measure_clock vec",
        "Clock frequency of HDMI": "vcgencmd measure_clock hdmi",
        "Clock frequency of Display Peripheral Interface": "vcgencmd measure_clock dpi",
        "Current voltage used by VC4 core": "vcgencmd measure_volts core",
        "Current voltage used by SDRAM core": "vcgencmd measure_volts sdram_c",
        "Current voltage used by SDRAM I/O": "vcgencmd measure_volts sdram_i",
        "Current voltage used by SDRAM Phy": "vcgencmd measure_volts sdram_p",
        "Amount of memory allocated to the ARM cores": "vcgencmd get_mem arm",
        "Amount of memory allocated to VC4": "vcgencmd get_mem gpu",
        "Out Of Memory events occuring in the VC4 memory": "vcgencmd mem_oom",
        "Power status of Main LCD": "vcgencmd display_power -1 0",
        "Power status of Secondary LCD": "vcgencmd display_power -1 1",
        "Power status of HDMI 0": "vcgencmd display_power -1 2",
        "Power status of Composite": "vcgencmd display_power -1 3",
        "Power status of HDMI 1": "vcgencmd display_power -1 7",
        "Resolution & color depth of displays": "vcgencmd get_lcd_info",
    }

    def run(self):
        StdoutFormat.print_header("SYSTEM HEALTH CHECK")
        print(f"Current time: {strftime('%a, %d %b %Y %I:%M:%S %p %Z')}")
        print("")

        self.print_raspberry_pi_device_info()
        print("")

        t = StdoutTable()
        t.add_section("pi-top Devices", self.get_pt_devices_status())
        t.print()

        t = StdoutTable()
        t.add_section("System Information", self.get_system_information())
        t.add_section(
            "Interfaces (via raspi-config)",
            self.get_raspi_config_settings(self.RASPI_CONFIG_INTERFACES),
        )
        t.add_section(
            "Boot Settings (via raspi-config)",
            self.get_raspi_config_settings(self.RASPI_CONFIG_BOOT_SETTINGS),
        )
        t.add_section(
            "Misc (via raspi-config)",
            self.get_raspi_config_settings(self.RASPI_CONFIG_MISC),
        )
        t.print()
        print("")

        StdoutFormat.print_section("Network Settings")
        self.print_network_settings()
        print("")

        StdoutFormat.print_section("pi-top Software Information")
        pt_sw = PitopSoftware()
        StdoutFormat.print_subsection("pi-top Systemd Services")
        pt_sw.print_pt_systemd_status()

        StdoutFormat.print_subsection("pi-top Installed Software")
        StdoutTable().print_data(pt_sw.get_pt_installed_software())

        StdoutFormat.print_subsection("APT Sources")
        pt_sw.print_apt_sources()
        print("")

        StdoutFormat.print_section("pi-top Device Information")
        try:
            hub = HubCommunication()
            hub.print_hub_registers()
        except Exception as e:
            print(f"{e}")

        StdoutFormat.print_subsection("Kernel Diagnostic Messages (dmesg)")
        self.print_dmesg()

    def get_pt_devices_status(self):
        data_arr = []
        for periph in pitop_peripherals():
            connection_status = f"[ {StdoutFormat.GREEN}{'✓' if periph.get('connected') else ' '}{StdoutFormat.ENDC} ]"
            device_name = f"{periph.get('name')}"
            if periph.get("fw_version"):
                device_name += f"(v{periph.get('fw_version')})"
            data_arr.append([connection_status, device_name])
        return data_arr

    def print_dmesg(self):
        try:
            dmesg = run_command("dmesg", timeout=10)
        except Exception:
            dmesg = "Error reading dmesg"

        for line in dmesg.split("\n"):
            StdoutFormat.print_line(line)

    def print_raspberry_pi_device_info(self):
        data_arr = [
            ("Device", f"{run_command('cat /proc/device-tree/model', timeout=2)}"),
            ("Architecture", f"{uname().machine}"),
        ]

        t = StdoutTable()
        t.add_section("Raspberry Pi Device Information", data_arr)
        t.print()

        if device_type() == DeviceName.pi_top_4.value:
            StdoutFormat.print_subsection("Raspberry Pi 4 Bootloader Configuration")
            eeprom_info = run_command("sudo rpi-eeprom-config", timeout=5, check=False)
            print(f"{eeprom_info.strip()}")
            StdoutFormat.print_subsection("Raspberry Pi 4 EEPROM Information")
            eeprom_info = run_command("sudo rpi-eeprom-update", timeout=5, check=False)
            print(f"{eeprom_info.strip()}")
            StdoutFormat.print_subsection("VideoCore GPU Detailed Configuration")
            self.print_vcgenmod_settings()

    def print_vcgenmod_settings(self):
        t = StdoutTable()
        data_arr = []
        for description, command in self.VCGENMOD_SETTINGS.items():
            try:
                result = run_command(command, timeout=1)
                result = result.strip().replace("\n", "; ")
            except Exception:
                result = "Error retrieving information"
            data_arr.append([description, result])
        t.print_data(data_arr)

    def get_system_information(self):
        sys_info_arr = [
            ("Debian Version", get_debian_version()),
            ("Kernel Version", get_uname_release()),
            ("Kernel Release", get_uname_version()),
        ]

        data = get_pitopOS_info()
        for k, v in data.__dict__.items():
            sys_info_arr.append((k.replace("_", " ").title(), v))

        return sys_info_arr

    def get_raspi_config_setting_value(self, setting):
        try:
            return run_command(f"raspi-config nonint {setting}", timeout=5).strip()
        except Exception:
            return "Error getting setting"

    def get_raspi_config_settings(self, raspi_config_settings_dict):
        data_to_print = []

        for setting, setting_dict in raspi_config_settings_dict.items():
            setting_value = self.get_raspi_config_setting_value(setting)
            if setting_dict.get("conversion_func"):
                setting_value = setting_dict.get("conversion_func")(setting_value)
            setting_value = str(setting_value)

            data_to_print.append((setting_dict.get("description"), setting_value))

        return data_to_print

    def print_network_settings(self):
        def print_interface_info(interface_name):
            iface_info = ifaddresses(interface_name)
            # get network layer, ipv4 and ipv6 info for the provided interface
            for netiface_enum, address_family in self.NETWORK_ENUM_LOOKUP.items():
                interface_info = iface_info.get(netiface_enum)
                if not interface_info:
                    continue

                # Print interface information for a particular "address family"
                StdoutFormat.print_line(f"{address_family}")
                for address_number, address_info in enumerate(interface_info):
                    # An interface can have more than one address associated to it
                    if len(interface_info) > 1:
                        StdoutFormat.print_line(
                            f"Subaddress #{address_number + 1}", level=2
                        )
                    # Print interface attributes & values for the address
                    data_arr = []
                    for addr_attribute, addr_attribute_value in address_info.items():
                        data_arr.append((addr_attribute, addr_attribute_value))

                    t = StdoutTable(indent_level=3 if len(interface_info) > 1 else 2)
                    t.print_data(data_arr)

        t = StdoutTable()
        t.print_data(
            [
                ("Hostname", f"{self.get_raspi_config_setting_value('get_hostname')}"),
                (
                    "WiFi Country",
                    f"{self.get_raspi_config_setting_value('get_wifi_country')}",
                ),
            ]
        )

        interfaces_list = interfaces()
        # Omit loopback interface
        if "lo" in interfaces_list:
            interfaces_list.remove("lo")
        for iface in interfaces_list:
            StdoutFormat.print_subsection(f"Interface: {iface}")
            print_interface_info(iface)



import re

# def bytes2human(n):
#     """
#     >>> bytes2human(10000)
#     '9K'
#     >>> bytes2human(100001221)
#     '95M'
#     """
#     symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
#     prefix = {}
#     for i, s in enumerate(symbols):
#         prefix[s] = 1 << (i + 1) * 10
#     for s in reversed(symbols):
#         if n >= prefix[s]:
#             value = int(float(n) / prefix[s])
#             return '%s%s' % (value, s)
#     return "%sB" % n


def bytes2human(n, fmt="{0:0.2f}"):
    symbols = [
        ("YB", 2 ** 80),
        ("ZB", 2 ** 70),
        ("EB", 2 ** 60),
        ("PB", 2 ** 50),
        ("TB", 2 ** 40),
        ("GB", 2 ** 30),
        ("MB", 2 ** 20),
        ("KB", 2 ** 10),
        ("B", 2 ** 0),
    ]

    for suffix, v in symbols:
        if n >= v:
            value = float(n) / v
            return fmt.format(value) + suffix
    return "{0}B".format(n)


def strip_whitespace(line):
    return "".join(line.split())


def is_line_commented(line_to_check, delimiter="#"):
    stripped_line = strip_whitespace(line_to_check)
    return stripped_line.startswith(delimiter)


def get_commented_line(line_to_change, delimiter="#"):
    stripped_line = strip_whitespace(line_to_change)
    commented_line = delimiter + stripped_line
    return commented_line


def get_uncommented_line(line_to_change, delimiter="#"):
    return line_to_change.replace(delimiter, "")


def remove_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


def is_url(text):
    url_regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        # domain...
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    return re.match(url_regex, text) is not None



import logging
from os import environ
from shlex import split
from subprocess import DEVNULL, CalledProcessError, Popen, TimeoutExpired, run

from pitop.common.current_session_info import get_first_display

logger = logging.getLogger(__name__)


def __get_env():
    env_plus_display = environ.copy()
    first_display = get_first_display()
    if first_display is not None:
        env_plus_display["DISPLAY"] = first_display
    return env_plus_display


def run_command_background(command_str: str, print_output=False) -> Popen:
    logger.debug("Function: run_command_background(command_str=%s)" % command_str)

    return Popen(
        split(command_str),
        env=__get_env(),
        stderr=None if print_output else DEVNULL,
        stdout=None if print_output else DEVNULL,
    )


def run_command(
    command_str: str,
    timeout: int,
    check: bool = True,
    capture_output: bool = True,
    log_errors: bool = True,
    lower_priority: bool = False,
) -> str:
    logger.debug(
        f"Function: run_command(command_str={command_str}, timeout={timeout}, check={check}, capture_output={capture_output}, \
         log_errors={log_errors}, lower_priority={lower_priority})"
    )

    resp_stdout = None

    if lower_priority:
        command_str = "nice -n 10 " + command_str

    try:
        resp = run(
            split(command_str),
            check=check,
            capture_output=capture_output,
            timeout=timeout,
            env=__get_env(),
        )

    except (CalledProcessError, TimeoutExpired) as e:
        if log_errors:
            logger.error(str(e))
        raise e
    except Exception as e:
        if log_errors:
            logger.error(str(e))
        return ""

    if capture_output:
        resp_stdout = str(resp.stdout, "utf8")
        resp_stderr = str(resp.stderr, "utf8")

        logger.debug(
            f"run_command("
            f"command_str='{command_str}', "
            f"timeout={timeout}, "
            f"check='{check}', "
            f"capture_output='{capture_output}'"
            f") stdout:\n{resp_stdout}"
        )
        logger.debug(
            f"run_command("
            f"command_str='{command_str}', "
            f"timeout={timeout}, "
            f"check='{check}', "
            f"capture_output='{capture_output}'"
            f") stderr:\n{resp_stderr}"
        )

    if not check:
        logger.debug(
            f"run_command("
            f"command_str='{command_str}', "
            f"timeout={timeout}, "
            f"check='{check}', "
            f"capture_output='{capture_output}'"
            f") exit code: {resp.returncode}"
        )

    return resp_stdout



from concurrent.futures import ThreadPoolExecutor
from inspect import signature

from pitop.core import ImageFunctions

from .capture_action_base import CaptureActionBase


class GenericAction(CaptureActionBase):
    """Class that executes the provided :data:`callback` whenever a camera
    frame is processed.

    :param function callback_on_frame: A callback function that will be called with each new frame as the first argument.
    :param int frame_interval: The callback will run every frame_interval frames, decreasing the frame rate of processing.
    """

    def __init__(self, callback_on_frame, frame_interval, format="PIL"):
        self.__generic_action_callback = callback_on_frame
        self.__event_executor = ThreadPoolExecutor()
        self.__frame_interval = frame_interval
        self.__elapsed_frames = 0
        self.__format = format

        callback_signature = signature(callback_on_frame)
        self.callback_has_argument = len(callback_signature.parameters) > 0

    def __del__(self):
        self.stop()

    def process(self, frame):
        if isinstance(self.__format, str) and self.__format.lower() == "opencv":
            frame = ImageFunctions.convert(frame, format="opencv")

        if self.__elapsed_frames % self.__frame_interval == 0:
            if self.callback_has_argument:
                self.__event_executor.submit(self.__generic_action_callback, frame)
            else:
                self.__event_executor.submit(self.__generic_action_callback)
        self.__elapsed_frames = (self.__elapsed_frames + 1) % self.__frame_interval

    def stop(self):
        if self.__event_executor is not None:
            self.__event_executor.shutdown()
            self.__event_executor = None



# classes to perform actions with frames
from .capture_action_base import CaptureActionBase

# enum with available capture actions
from .capture_actions import CaptureActions
from .generic_action import GenericAction
from .motion_detector import MotionDetector
from .store_frame import StoreFrame
from .video_capture import VideoCapture



from flask import Blueprint, g

from pitop.labs.web.blueprints.controller import ControllerBlueprint

from .helpers import calculate_pan_tilt_angle, calculate_velocity_twist


def drive_handler(drive, data):
    velocity_twist = calculate_velocity_twist(data)
    linear = velocity_twist.get("linear", 0)
    angular = velocity_twist.get("angular", 0)

    drive.robot_move(linear, angular)


def pan_tilt_handler(pan_tilt, data):
    angle = calculate_pan_tilt_angle(data)

    pan_tilt.pan_servo.target_angle = angle.get("z")
    pan_tilt.tilt_servo.target_angle = angle.get("y")


class RoverControllerBlueprint(Blueprint):
    def __init__(
        self, drive=None, pan_tilt=None, get_frame=None, message_handlers={}, **kwargs
    ):
        Blueprint.__init__(
            self, "rover", __name__, template_folder="templates", **kwargs
        )

        passed_left_joystick_handler = message_handlers.get("left_joystick") is not None

        # set show_left_joystick every request for use in base-rover.html
        @self.before_app_request
        def set_show_left_joystick():
            g.show_left_joystick = passed_left_joystick_handler or pan_tilt is not None

        if message_handlers.get("left_joystick") is None:

            def left_joystick(data):
                return pan_tilt_handler(pan_tilt, data)

            message_handlers["left_joystick"] = left_joystick

        if message_handlers.get("right_joystick") is None:

            def right_joystick(data):
                return drive_handler(drive, data)

            message_handlers["right_joystick"] = right_joystick

        self.controller_blueprint = ControllerBlueprint(
            get_frame=get_frame, message_handlers=message_handlers
        )

    def register(self, app, options, *args, **kwargs):
        app.register_blueprint(self.controller_blueprint, **options)
        Blueprint.register(self, app, options, *args, **kwargs)

    def broadcast(self, message):
        self.controller_blueprint.broadcast(message)



import importlib
from json import dump, dumps, load


class Recreatable:
    """Represents an object that keeps track of a set of parameters that will
    allow to be recreate it in the future.

    The values for each key provided in the :param:`config_dict`
    parameter can be a constant value or a reference to a function,
    which will be evaluated when the object configuration is requested.
    This is useful for cases where a parameter changes its value on
    runtime.
    """

    def __init__(self, config_dict=None):
        if config_dict is None:
            config_dict = dict()

        if not isinstance(config_dict, dict):
            raise TypeError("Argument must be a dictionary")

        config_dict.update(
            {
                "classname": self.__class__.__name__,
                "module": self.__module__,
                "version": "0.17.0",
            }
        )
        for k, v in config_dict.items():
            config_dict.update({k: v})

        self._config = config_dict

    @classmethod
    def from_config(cls, config_dict):
        """Creates an instance of a Recreatable object with parameters in the
        provided dictionary."""
        _ = config_dict.pop("version", None)
        cls_name = config_dict.pop("classname")
        module_name = config_dict.pop("module")

        message = f"Recreating {cls_name} (from {module_name})"
        message += f" with {config_dict}" if config_dict else ""
        print(message)

        main_cls = cls.import_class(module_name, cls_name)
        return main_cls(**config_dict)

    @classmethod
    def from_file(cls, path):
        """Creates an instance of an object using the JSON file from the
        provided path."""
        print(f"Loading configuration from {path}")
        with open(path) as json_file:
            config_dict = load(json_file)
        return cls.from_config(config_dict)

    def save_config(self, path):
        """Stores the set of parameters to recreate an object in a JSON
        file."""
        print(f"Storing configuration in {path}")
        with open(path, "w") as writer:
            dump(self.config, writer, indent=4)

    @property
    def config(self):
        """Returns a dictionary with the set of parameters that can be used to
        recreate an object."""
        cfg = {}
        for k, v in self._config.items():
            cfg[k] = v() if callable(v) else v
        return cfg

    @staticmethod
    def import_class(module_name, class_name):
        """Imports a class given a module and a class name."""
        module = importlib.import_module(module_name)
        return getattr(module, class_name)

    def print_config(self):
        print(dumps(self.config, indent=4))



import logging
from glob import glob
from os import environ

logger = logging.getLogger(__name__)


def get_current_user():
    """Returns the name of the user that invoked this function.

    Returns:
            user (str): String representing the user
    """
    if environ.get("SUDO_USER"):
        return environ.get("SUDO_USER")
    elif environ.get("USER"):
        return environ.get("USER")
    else:
        return get_user_using_first_display()


def get_list_of_displays() -> int:
    display_file_prefix = "/tmp/.X11-unix/X"
    return [
        f.replace(display_file_prefix, ":")
        for f in glob(display_file_prefix + "[0-9]*")
    ]


def get_first_display() -> int:
    displays = get_list_of_displays()
    first_display = displays[0] if len(displays) > 0 else None
    logger.debug("First display is: {}".format(first_display))
    return first_display


def get_user_using_display(display_no):
    """Returns the name of the user that is currently using the defined
    display.

    Returns:
            user (str): String representing the user
    """
    from pitop.common.command_runner import run_command

    user = None
    for line in run_command("who", timeout=1).split("\n"):
        if "(%s)" % display_no in line:
            fields = line.split(" ")
            if len(fields) > 1:
                user = fields[0]
                break
    return user


def get_user_using_first_display():
    """Returns the name of the user that is currently using the first available
    display.

    This function is useful when targeting a particular active user by something that is running
    as something different to the current user, where `get_current_user()` would be incorrect.

    For example, with a system service, `get_current_user()` would return "root" (from the USER
    environment variable), where the active user (e.g. "pi") is actually wanted.

            Returns:
                    user (str): String representing the user
    """
    return get_user_using_display(get_first_display())



from flask import Blueprint


class BaseBlueprint(Blueprint):
    def __init__(self, name="base", **kwargs):
        Blueprint.__init__(
            self,
            name,
            __name__,
            template_folder="templates",
            static_folder="base",
            **kwargs
        )



from numpy import append, cumsum, delete, insert, shape


def running_mean(old_array, new_value):
    """Calculates a running mean from either a 1d numpy array of single
    measurements or a 2d numpy array of 1d numpy array measurements.

    :param old_array: Stored historical values either as a 1d numpy array or a 2d numpy array
    :param new_value: New value to add. For a 1d array this will be a single value. For a 2d Nxm numpy array, this
    will be a 1d numpy array that has m values, where N is the number of samples to average over.
    :return: tuple where index 0 is the new array and index 1 is the new mean.
    """

    def calculate_mean(x, N):
        cum_sum = cumsum(insert(x, 0, 0, axis=0), axis=0)
        return (cum_sum[N:] - cum_sum[:-N]) / float(N)

    array_shape = old_array.shape
    if old_array.ndim == 2:
        new_value = new_value.reshape(1, array_shape[1])
        new_array = append(delete(old_array, 0, axis=0), new_value, axis=0)
    else:
        new_array = append(delete(old_array, 0, axis=0), new_value)

    new_mean = calculate_mean(new_array, shape(new_array)[0])[0]
    return new_array, new_mean


def remap_range(old_value, old_range, new_range, invert_direction=False):
    old_min = old_range[0]
    old_max = old_range[1]
    old_range_magnitude = abs(old_max - old_min)
    new_min = new_range[0]
    new_max = new_range[1]
    new_range_magnitude = abs(new_max - new_min)

    if not invert_direction:
        new_value = (
            ((old_value - old_min) * new_range_magnitude) / old_range_magnitude
        ) + new_min
    else:
        new_value = new_max - (
            ((old_value - old_min) * new_range_magnitude) / old_range_magnitude
        )

    return new_value



# camera devices
from .cameras import CameraTypes, FileSystemCamera, UsbCamera

# classes to perform actions with frames
from .capture_actions import (
    CaptureActionBase,
    CaptureActions,
    GenericAction,
    MotionDetector,
    StoreFrame,
    VideoCapture,
)
from .frame_handler import FrameHandler



# coding=utf-8
# pynput
# Copyright (C) 2015-2018 Moses Palmér
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""General utility functions and classes."""

# pylint: disable=R0903
# We implement minimal mixins

# pylint: disable=W0212
# We implement an internal API

import contextlib
import functools
import sys
import threading

import six
from six.moves import queue


class AbstractListener(threading.Thread):
    """A class implementing the basic behaviour for event listeners.

    Instances of this class can be used as context managers. This is equivalent
    to the following code::

        listener.start()
        listener.wait()
        try:
            with_statements()
        finally:
            listener.stop()

    Actual implementations of this class must set the attribute ``_log``, which
    must be an instance of :class:`logging.Logger`.

    :param bool suppress: Whether to suppress events. Setting this to ``True``
        will prevent the input events from being passed to the rest of the
        system.

    :param kwargs: A mapping from callback attribute to callback handler. All
        handlers will be wrapped in a function reading the return value of the
        callback, and if it ``is False``, raising :class:`StopException`.

        Any callback that is falsy will be ignored.
    """

    class StopException(Exception):
        """If an event listener callback raises this exception, the current
        listener is stopped."""

        pass

    #: Exceptions that are handled outside of the emitter and should thus not be
    #: passed through the queue
    _HANDLED_EXCEPTIONS = tuple()

    def __init__(self, suppress=False, **kwargs):
        super(AbstractListener, self).__init__()

        def wrapper(f):
            def inner(*args):
                if f(*args) is False:
                    raise self.StopException()

            return inner

        self._suppress = suppress
        self._running = False
        self._thread = threading.current_thread()
        self._condition = threading.Condition()
        self._ready = False

        # Allow multiple calls to stop
        self._queue = queue.Queue(10)

        self.daemon = True

        for name, callback in kwargs.items():
            setattr(self, name, wrapper(callback or (lambda *a: None)))

    @property
    def suppress(self):
        """Whether to suppress events."""
        return self._suppress

    @property
    def running(self):
        """Whether the listener is currently running."""
        return self._running

    def stop(self):
        """Stops listening for events.

        When this method returns, no more events will be delivered. Once this
        method has been called, the listener instance cannot be used any more,
        since a listener is a :class:`threading.Thread`, and once stopped it
        cannot be restarted.

        To resume listening for event, a new listener must be created.
        """
        if self._running:
            self._running = False
            self._queue.put(None)
            self._stop_platform()

    def __enter__(self):
        self.start()
        self.wait()
        return self

    def __exit__(self, exc_type, value, traceback):
        self.stop()

    def wait(self):
        """Waits for this listener to become ready."""
        self._condition.acquire()
        while not self._ready:
            self._condition.wait()
        self._condition.release()

    def run(self):
        """The thread runner method."""
        self._running = True
        self._thread = threading.current_thread()
        self._run()

        # Make sure that the queue contains something
        self._queue.put(None)

    @classmethod
    def _emitter(cls, f):
        """A decorator to mark a method as the one emitting the callbacks.

        This decorator will wrap the method and catch exception. If a
        :class:`StopException` is caught, the listener will be stopped
        gracefully. If any other exception is caught, it will be propagated to
        the thread calling :meth:`join` and reraised there.
        """

        @functools.wraps(f)
        def inner(self, *args, **kwargs):
            # pylint: disable=W0702; we want to catch all exception
            try:
                return f(self, *args, **kwargs)
            except Exception as e:
                if not isinstance(e, self._HANDLED_EXCEPTIONS):
                    if not isinstance(e, AbstractListener.StopException):
                        self._log.exception("Unhandled exception in listener callback")
                    self._queue.put(
                        None if isinstance(e, cls.StopException) else sys.exc_info()
                    )
                    self.stop()
                raise
            # pylint: enable=W0702

        return inner

    def _mark_ready(self):
        """Marks this listener as ready to receive events.

        This method must be called from :meth:`_run`. :meth:`wait` will
        block until this method is called.
        """
        self._condition.acquire()
        self._ready = True
        self._condition.notify()
        self._condition.release()

    def _run(self):
        """The implementation of the :meth:`run` method.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def _stop_platform(self):
        """The implementation of the :meth:`stop` method.

        This is a platform dependent implementation.
        """
        raise NotImplementedError()

    def join(self, *args):
        super(AbstractListener, self).join(*args)

        # Reraise any exceptions
        try:
            exc_type, exc_value, exc_traceback = self._queue.get()
        except TypeError:
            return
        six.reraise(exc_type, exc_value, exc_traceback)


class NotifierMixin(object):
    """A mixin for notifiers of fake events.

    This mixin can be used for controllers on platforms where sending
    fake events does not cause a listener to receive a notification.
    """

    def _emit(self, action, *args):
        """Sends a notification to all registered listeners.

        This method will ensure that listeners that raise
        :class:`StopException` are stopped.

        :param str action: The name of the notification.

        :param args: The arguments to pass.
        """
        stopped = []
        for listener in self._listeners():
            try:
                getattr(listener, action)(*args)
            except listener.StopException:
                stopped.append(listener)
        for listener in stopped:
            listener.stop()

    @classmethod
    def _receiver(cls, listener_class):
        """A decorator to make a class able to receive fake events from a
        controller.

        This decorator will add the method ``_receive`` to the decorated class.

        This method is a context manager which ensures that all calls to
        :meth:`_emit` will invoke the named method in the listener instance
        while the block is active.
        """

        @contextlib.contextmanager
        def receive(self):
            """Executes a code block with this listener instance registered as
            a receiver of fake input events."""
            self._controller_class._add_listener(self)
            try:
                yield
            finally:
                self._controller_class._remove_listener(self)

        listener_class._receive = receive
        listener_class._controller_class = cls

        # Make sure this class has the necessary attributes
        if not hasattr(cls, "_listener_cache"):
            cls._listener_cache = set()
            cls._listener_lock = threading.Lock()

        return listener_class

    @classmethod
    def _listeners(cls):
        """Iterates over the set of running listeners.

        This method will quit without acquiring the lock if the set is
        empty, so there is potential for race conditions. This is an
        optimisation, since :class:`Controller` will need to call this
        method for every control event.
        """
        if not cls._listener_cache:
            return
        with cls._listener_lock:
            for listener in cls._listener_cache:
                yield listener

    @classmethod
    def _add_listener(cls, listener):
        """Adds a listener to the set of running listeners.

        :param listener: The listener for fake events.
        """
        with cls._listener_lock:
            cls._listener_cache.add(listener)

    @classmethod
    def _remove_listener(cls, listener):
        """Removes this listener from the set of running listeners.

        :param listener: The listener for fake events.
        """
        with cls._listener_lock:
            cls._listener_cache.remove(listener)
